# Involute Gears Generation Script
# by Marcin Wanczyk (dj_who)
# (c) 2011 LGPL


import FreeCAD
import FreeCADGui
import Part
import Draft
import MeshPart
import Mesh
import math
from PySide import QtGui, QtCore

App = FreeCAD
Gui = FreeCADGui


def proceed():
    try:
        compute()
    except Exception:
        hide()
        QtGui.QApplication.restoreOverrideCursor()


def compute():
    QtGui.QApplication.setOverrideCursor(QtCore.Qt.WaitCursor)

    if FreeCAD.ActiveDocument is None:
        FreeCAD.newDocument("Gear")

    oldDocumentObjects = App.ActiveDocument.Objects

    try:
        N = int(l1.text())
        p = float(l2.text())
        alfa = int(l3.text())
        y = float(l4.text())  # standard value y<1 for gear drives y>1 for Gear pumps
        m = (
            p / math.pi
        )  # standard value 0.06, 0.12, 0.25, 0.5, 1, 2, 4, 8, 16, 32, 60 (polish norm)
        c = float(l5.text()) * m  # standard value 0,1*m - 0,3*m
        j = float(l6.text()) * m  # standard value 0,015 - 0,04*m
        width = float(l7.text())  # gear width
    except ValueError:
        FreeCAD.Console.PrintError("Wrong input! Only numbers allowed.\n")

    # tooth height
    h = 2 * y * m + c

    # pitch diameter
    d = N * m

    # root diameter
    df = d - 2 * y * m - 2 * c  # df=d-2hf where and hf=y*m+c

    # addendum diameter
    da = d + 2 * y * m  # da=d+2ha where ha=y*m

    # base diameter for involute
    db = d * math.cos(math.radians(alfa))

    # Base circle
    baseCircle = FreeCAD.ActiveDocument.addObject("Part::FeaturePython", "BaseCircle")
    Draft._Circle(baseCircle)
    Draft._ViewProviderDraft(baseCircle.ViewObject)
    baseCircle.Radius = db / 2
    baseCircle.FirstAngle = 0.0
    baseCircle.LastAngle = 0.0

    # Root circle
    rootCircle = FreeCAD.ActiveDocument.addObject("Part::FeaturePython", "RootCircle")
    Draft._Circle(rootCircle)
    Draft._ViewProviderDraft(rootCircle.ViewObject)
    rootCircle.Radius = df / 2
    rootCircle.FirstAngle = 0.0
    rootCircle.LastAngle = 0.0

    # Addendum circle
    addendumCircle = FreeCAD.ActiveDocument.addObject("Part::FeaturePython", "AddendumCircle")
    Draft._Circle(addendumCircle)
    Draft._ViewProviderDraft(addendumCircle.ViewObject)
    addendumCircle.Radius = da / 2
    addendumCircle.FirstAngle = 0.0
    addendumCircle.LastAngle = 0.0

    # Pitch circle
    pitchCircle = FreeCAD.ActiveDocument.addObject("Part::FeaturePython", "PitchCircle")
    Draft._Circle(pitchCircle)
    Draft._ViewProviderDraft(pitchCircle.ViewObject)
    pitchCircle.Radius = d / 2
    pitchCircle.FirstAngle = 0.0
    pitchCircle.LastAngle = 0.0

    # ************ Calculating right sides of teeth
    # Involute of base circle
    involute = []
    involutee = []
    involutesav = []

    for t in range(0, 60, 1):
        x = db / 2 * (math.cos(math.radians(t)) + math.radians(t) * math.sin(math.radians(t)))
        y = db / 2 * (math.sin(math.radians(t)) - math.radians(t) * math.cos(math.radians(t)))
        involute.append(Part.Vertex(x, y, 0).Point)

    # ************ Drawing right sides of teeth
    involutesav.extend(involute)
    involutee.extend(involute)

    for angle in range(1, N + 1, 1):
        involuteobj = FreeCAD.ActiveDocument.addObject("Part::Feature", "InvoluteL" + str(angle))
        involutee.insert(0, (0, 0, 0))
        involuteshape = Part.makePolygon(involutee)
        involuteobj.Shape = involuteshape
        involutee = []
        for num in range(0, 60, 1):
            point = involute.pop()
            pointt = Part.Vertex(
                point.x * math.cos(math.radians(angle * 360 / N))
                - point.y * math.sin(math.radians(angle * 360 / N)),
                point.x * math.sin(math.radians(angle * 360 / N))
                + point.y * math.cos(math.radians(angle * 360 / N)),
                0,
            ).Point
            involutee.insert(0, pointt)
        involute.extend(involutesav)
    involutee = []

    # ************ Calculating difference between tooth spacing on BaseCircle and PitchCircle

    pc = App.ActiveDocument.getObject("PitchCircle")
    inv = App.ActiveDocument.getObject("InvoluteL1")
    cut = inv.Shape.cut(pc.Shape)
    #    FreeCAD.ActiveDocument.addObject("Part::Feature","CutInv").Shape=cut
    invPoint = cut.Vertexes[0].Point

    diff = invPoint.y * 2  # instead of making axial symmetry and calculating point distance.
    anglediff = 2 * math.asin(diff / d)

    # ************ Calculating left sides of teeth

    # ************ Inversing Involute
    for num in range(0, 60, 1):
        point = involute.pop()
        pointt = Part.Vertex(point.x, point.y * -1, 0).Point
        involutee.insert(0, pointt)
    involute.extend(involutee)
    involutee = []

    # Normal tooth size calculated as: 0,5*  p    -      j                       j = m * 0,1 below are calculations
    #                                 0,5*  p    -      m      * 0,1
    #                                 0,5*  p    -     p   /pi * 0,1
    #                                 0,5*360/N  - ((360/N)/pi)* 0,1
    #                                 0,5*360/N  - (360/N)*((1/pi)*0,1)         j = (p/pi)*0,1
    #                                 0,5*360/N  - (360/N)*((p/pi)*0,1)/p
    #                                 0,5*360/N  - (360/N)*(    j     )/p
    for num in range(0, 60, 1):
        point = involute.pop()
        pointt = Part.Vertex(
            point.x * math.cos(math.radians(180 / N - (360 / N) * (j / p)) + anglediff)
            - point.y * math.sin(math.radians(180 / N - (360 / N) * (j / p)) + anglediff),
            point.x * math.sin(math.radians(180 / N - (360 / N) * (j / p)) + anglediff)
            + point.y * math.cos(math.radians(180 / N - (360 / N) * (j / p)) + anglediff),
            0,
        ).Point
        involutee.insert(0, pointt)
    involute.extend(involutee)
    involutesav = []
    involutesav.extend(involute)

    # ************ Drawing left sides of teeth
    for angle in range(1, N + 1, 1):
        involuteobj = FreeCAD.ActiveDocument.addObject("Part::Feature", "InvoluteR" + str(angle))
        involutee.insert(0, (0, 0, 0))
        involuteshape = Part.makePolygon(involutee)
        involuteobj.Shape = involuteshape
        involutee = []
        for num in range(0, 60, 1):
            point = involute.pop()
            pointt = Part.Vertex(
                point.x * math.cos(math.radians(angle * 360 / N))
                - point.y * math.sin(math.radians(angle * 360 / N)),
                point.x * math.sin(math.radians(angle * 360 / N))
                + point.y * math.cos(math.radians(angle * 360 / N)),
                0,
            ).Point
            involutee.insert(0, pointt)
        involute.extend(involutesav)

    Gui.SendMsgToActiveView("ViewFit")

    # ************ Forming teeth

    cutCircle = FreeCAD.ActiveDocument.addObject("Part::FeaturePython", "CutCircle")
    Draft._Circle(cutCircle)
    Draft._ViewProviderDraft(cutCircle.ViewObject)
    cutCircle.Radius = da  # da because must be bigger than addendumCircle and bigger than whole construction da is right for this but it not has to be.
    cutCircle.FirstAngle = 0.0
    cutCircle.LastAngle = 0.0

    cutTool = cutCircle.Shape.cut(addendumCircle.Shape)
    # cutshape = Part.show(cutTool)

    gearShape = rootCircle.Shape

    for invNum in range(1, N + 1, 1):
        invL = App.ActiveDocument.getObject("InvoluteL" + str(invNum))
        invR = App.ActiveDocument.getObject("InvoluteR" + str(invNum))
        cutL = invL.Shape.cut(cutTool)
        cutR = invR.Shape.cut(cutTool)
        pointL = cutL.Vertexes.pop().Point
        pointR = cutR.Vertexes.pop().Point
        faceEdge = Part.makeLine(pointL, pointR)

        toothWhole = cutL.fuse(cutR)
        toothWhole = toothWhole.fuse(faceEdge)
        toothWire = Part.Wire(toothWhole.Edges)
        toothShape = Part.Face(toothWire)
        #        tooth = App.ActiveDocument.addObject("Part::Feature", "Tooth" +str(invNum))
        #        tooth.Shape=toothShape
        gearShape = gearShape.fuse(toothShape)

    for o in App.ActiveDocument.Objects:
        if oldDocumentObjects.count(o) == 0:
            App.ActiveDocument.removeObject(o.Name)

    gearFlat = App.ActiveDocument.addObject("Part::Feature", "GearFlat")
    gearFlat.Shape = gearShape
    Gui.ActiveDocument.getObject(gearFlat.Name).Visibility = False

    gear = App.ActiveDocument.addObject("Part::Extrusion", "Gear3D")
    gear.Base = gearFlat
    gear.Dir = (0, 0, width)
    App.ActiveDocument.recompute()

    if c1.isChecked():
        gearMesh = App.ActiveDocument.addObject("Mesh::Feature", "Gear3D-mesh")

        faces = []
        triangles = gear.Shape.tessellate(
            1
        )  # the number represents the precision of the tessellation)
        for tri in triangles[1]:
            face = []
            for i in range(3):
                vindex = tri[i]
                face.append(triangles[0][vindex])
            faces.append(face)
        mesh = Mesh.Mesh(faces)

        gearMesh.Mesh = mesh
        App.ActiveDocument.removeObject(gear.Name)
        App.ActiveDocument.removeObject(gearFlat.Name)

    App.ActiveDocument.recompute()
    Gui.SendMsgToActiveView("ViewFit")

    QtGui.QApplication.restoreOverrideCursor()

    hide()


def hide():
    dialog.hide()


dialog = QtGui.QDialog()
dialog.resize(200, 450)
dialog.setWindowTitle("Gear")
la = QtGui.QVBoxLayout(dialog)
t1 = QtGui.QLabel("Number of teeth (N)")
la.addWidget(t1)
l1 = QtGui.QLineEdit()
l1.setText("16")
la.addWidget(l1)
t2 = QtGui.QLabel("Circular pitch (p)")
la.addWidget(t2)
l2 = QtGui.QLineEdit()
l2.setText("1.65")
la.addWidget(l2)
t3 = QtGui.QLabel("Pressure angle (alfa)")
la.addWidget(t3)
l3 = QtGui.QLineEdit()
l3.setText("20")
la.addWidget(l3)
t4 = QtGui.QLabel("Tooth height factor (y)")
la.addWidget(t4)
l4 = QtGui.QLineEdit()
l4.setText("1.0")
la.addWidget(l4)
t5 = QtGui.QLabel("Tooth clearance (c)")
la.addWidget(t5)
l5 = QtGui.QLineEdit()
l5.setText("0.1")
la.addWidget(l5)
t6 = QtGui.QLabel("Tooth lateral clearance (j)")
la.addWidget(t6)
l6 = QtGui.QLineEdit()
l6.setText("0.04")
la.addWidget(l6)
t7 = QtGui.QLabel("Gear width")
la.addWidget(t7)
l7 = QtGui.QLineEdit()
l7.setText("6.0")
la.addWidget(l7)
c1 = QtGui.QCheckBox("Create as a Mesh")
la.addWidget(c1)
e1 = QtGui.QLabel("(for faster rendering)")
commentFont = QtGui.QFont("Times", 8, True)
e1.setFont(commentFont)
la.addWidget(e1)

okbox = QtGui.QDialogButtonBox(dialog)
okbox.setOrientation(QtCore.Qt.Horizontal)
okbox.setStandardButtons(QtGui.QDialogButtonBox.Cancel | QtGui.QDialogButtonBox.Ok)
la.addWidget(okbox)
QtCore.QObject.connect(okbox, QtCore.SIGNAL("accepted()"), proceed)
QtCore.QObject.connect(okbox, QtCore.SIGNAL("rejected()"), hide)
QtCore.QMetaObject.connectSlotsByName(dialog)
dialog.show()



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Jakub Michalski <jakub.j.michalski[at]gmail.com>         *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM calculix amplitude"
__author__ = "Jakub Michalski"
__url__ = "https://www.freecad.org"


def write_amplitude(f, ccxwriter):

    # write amplitude definitions for all analysis features that use them

    def write_obj_amplitude(obj):
        if obj.EnableAmplitude:
            f.write(f"*AMPLITUDE, NAME={obj.Name}\n")
            for value in obj.AmplitudeValues:
                f.write(f"{value}\n")
            f.write("\n")

    constraint_lists = [
        ccxwriter.member.cons_force,
        ccxwriter.member.cons_pressure,
        ccxwriter.member.cons_displacement,
        ccxwriter.member.cons_heatflux,
        ccxwriter.member.cons_temperature,
        ccxwriter.member.cons_bodyheatsource,
        ccxwriter.member.cons_rigidbody,
        ccxwriter.member.cons_initialtemperature,
    ]

    for constraint_list in constraint_lists:
        for entry in constraint_list:
            write_obj_amplitude(entry["Object"])



# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileCopyrightText: 2023 FreeCAD Project Association
# SPDX-FileNotice: Part of the AddonManager.

################################################################################
#                                                                              #
#   This addon is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   This addon is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with this addon. If not, see https://www.gnu.org/licenses    #
#                                                                              #
################################################################################

"""Classes for working with Addon metadata, as documented at
https://wiki.FreeCAD.org/Package_metadata"""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import IntEnum, auto
from typing import Tuple, Dict, List, Optional

from addonmanager_licenses import get_license_manager

try:
    # If this system provides a secure parser, use that:
    import defusedxml.ElementTree as ET
except ImportError:
    # Otherwise fall back to the Python standard parser
    import xml.etree.ElementTree as ET


@dataclass
class Contact:
    name: str
    email: str = ""


@dataclass
class License:
    name: str
    file: str = ""


class UrlType(IntEnum):
    bugtracker = 0
    discussion = auto()
    documentation = auto()
    readme = auto()
    repository = auto()
    website = auto()

    def __str__(self):
        return f"{self.name}"


@dataclass
class Url:
    location: str
    type: UrlType
    branch: str = ""


class Version:
    """Provide a more useful representation of Version information"""

    def __init__(self, from_string: str = None, from_list=None):
        """If from_string is a string, it is parsed to get the version. If from_list
        exists (and no string was provided), it is treated as a version list of
        [major:int, minor:int, patch:int, pre:str]"""
        self.version_as_list = [0, 0, 0, ""]
        if from_string is not None:
            self._init_from_string(from_string)
        elif from_list is not None:
            self._init_from_list(from_list)

    def _init_from_string(self, from_string: str):
        """Find the first digit in the given string, and send that substring off for
        parsing."""
        counter = 0
        for char in from_string:
            if char.isdigit():
                break
            counter += 1
        self._parse_string_to_tuple(from_string[counter:])

    def _init_from_list(self, from_list):
        for index, element in enumerate(from_list):
            if index < 3:
                self.version_as_list[index] = int(element)
            elif index == 3:
                self.version_as_list[index] = str(element)
            else:
                break

    def _parse_string_to_tuple(self, from_string: str):
        """We hand-parse only simple version strings, of the form 1.2.3suffix -- only
        the first digit is required."""
        splitter = from_string.split(".", 2)
        counter = 0
        for component in splitter:
            try:
                self.version_as_list[counter] = int(component)
                counter += 1
            except ValueError:
                if counter == 0:
                    raise ValueError(f"Invalid version string {from_string}")
                number, text = self._parse_final_entry(component)
                self.version_as_list[counter] = number
                self.version_as_list[3] = text

    @staticmethod
    def _parse_final_entry(final_string: str) -> Tuple[int, str]:
        """The last value is permitted to contain both a number and a word, and needs
        to be split"""
        digits = ""
        for c in final_string:
            if c.isdigit():
                digits += c
            else:
                break
        return int(digits), final_string[len(digits) :]

    def __repr__(self) -> str:
        v = self.version_as_list
        return f"{v[0]}.{v[1]}.{v[2]} {v[3]}"

    def __eq__(self, other) -> bool:
        return self.version_as_list == other.version_as_list

    def __ne__(self, other) -> bool:
        return not (self == other)

    def __lt__(self, other) -> bool:
        for a, b in zip(self.version_as_list, other.version_as_list):
            if a != b:
                return a < b
        return False

    def __gt__(self, other) -> bool:
        if self.version_as_list == other.version_as_list:
            return False
        return not (self < other)

    def __ge__(self, other) -> bool:
        return self > other or self == other

    def __le__(self, other) -> bool:
        return self < other or self == other


class DependencyType(IntEnum):
    automatic = 0
    internal = auto()
    addon = auto()
    python = auto()

    def __str__(self):
        return f"{self.name}"


@dataclass
class Dependency:
    package: str
    version_lt: str = ""
    version_lte: str = ""
    version_eq: str = ""
    version_gte: str = ""
    version_gt: str = ""
    condition: str = ""
    optional: bool = False
    dependency_type: DependencyType = DependencyType.automatic


@dataclass
class GenericMetadata:
    """Used to store unrecognized elements"""

    contents: str = ""
    attributes: Dict[str, str] = field(default_factory=dict)


@dataclass
class Metadata:
    """A pure-python implementation of the Addon Manager's Metadata handling class"""

    name: str = ""
    version: Version = None
    date: str = ""
    description: str = ""
    type: str = ""
    maintainer: List[Contact] = field(default_factory=list)
    license: List[License] = field(default_factory=list)
    url: List[Url] = field(default_factory=list)
    author: List[Contact] = field(default_factory=list)
    depend: List[Dependency] = field(default_factory=list)
    conflict: List[Dependency] = field(default_factory=list)
    replace: List[Dependency] = field(default_factory=list)
    tag: List[str] = field(default_factory=list)
    icon: str = ""
    classname: str = ""
    subdirectory: str = ""
    file: List[str] = field(default_factory=list)
    freecadmin: Version = None
    freecadmax: Version = None
    pythonmin: Version = None
    content: Dict[str, List[Metadata]] = field(default_factory=dict)  # Recursive def.


def get_first_supported_freecad_version(metadata: Metadata) -> Optional[Version]:
    """Look through all content items of this metadata element and determine what the
    first version of freecad that ANY of the items support is. For example, if it
    contains several workbenches, some of which require v0.20, and some 0.21, then
    0.20 is returned. Returns None if freecadmin is unset by any part of this object."""

    current_earliest = metadata.freecadmin if metadata.freecadmin is not None else None
    for content_class in metadata.content.values():
        for content_item in content_class:
            content_first = get_first_supported_freecad_version(content_item)
            if content_first is not None:
                if current_earliest is None:
                    current_earliest = content_first
                else:
                    current_earliest = min(current_earliest, content_first)

    return current_earliest


def get_branch_from_metadata(metadata: Metadata) -> str:
    for url in metadata.url:
        if url.type == UrlType.repository:
            return url.branch
    return "master"  # Legacy default


def get_repo_url_from_metadata(metadata: Metadata) -> str:
    for url in metadata.url:
        if url.type == UrlType.repository:
            return url.location
    return ""


class MetadataReader:
    """Read metadata XML data and construct a Metadata object"""

    @staticmethod
    def from_file(filename: str) -> Metadata:
        """A convenience function for loading the Metadata from a file"""
        with open(filename, "rb") as f:
            data = f.read()
            return MetadataReader.from_bytes(data)

    @staticmethod
    def from_bytes(data: bytes) -> Metadata:
        """Read XML data from bytes and use it to construct Metadata"""
        element_tree = ET.fromstring(data)
        return MetadataReader._process_element_tree(element_tree)

    @staticmethod
    def _process_element_tree(root: ET.Element) -> Metadata:
        """Parse an element tree and convert it into a Metadata object"""
        namespace = MetadataReader._determine_namespace(root)
        return MetadataReader._create_node(namespace, root)

    @staticmethod
    def _determine_namespace(root: ET.Element) -> str:
        if root.tag.endswith("package"):
            return root.tag[: -len("package")]
        raise RuntimeError("No 'package' element found in metadata file")

    @staticmethod
    def _parse_child_element(namespace: str, child: ET.Element, metadata: Metadata):
        """Figure out what sort of metadata child represents, and add it to the
        metadata object."""

        tag = child.tag[len(namespace) :]
        if tag in ["name", "date", "description", "type", "icon", "classname", "subdirectory"]:
            # Text-only elements
            metadata.__dict__[tag] = child.text
        elif tag in ["version", "freecadmin", "freecadmax", "pythonmin"]:
            try:
                metadata.__dict__[tag] = Version(from_string=child.text)
            except ValueError:
                print(
                    f"Invalid version specified for tag {tag} in addon {metadata.name}: {child.text}"
                )
                metadata.__dict__[tag] = Version(from_list=[0, 0, 0])
        elif tag in ["tag", "file"]:
            # Lists of strings
            if child.text:
                metadata.__dict__[tag].append(child.text)
        elif tag in ["maintainer", "author"]:
            # Lists of contacts
            metadata.__dict__[tag].append(MetadataReader._parse_contact(child))
        elif tag == "license":
            # List of licenses
            metadata.license.append(MetadataReader._parse_license(child, metadata.name))
        elif tag == "url":
            # List of urls
            metadata.url.append(MetadataReader._parse_url(child))
        elif tag in ["depend", "conflict", "replace"]:
            # Lists of dependencies
            metadata.__dict__[tag].append(MetadataReader._parse_dependency(child))
        elif tag == "content":
            MetadataReader._parse_content(namespace, metadata, child)

    @staticmethod
    def _parse_contact(child: ET.Element) -> Contact:
        email = child.attrib["email"] if "email" in child.attrib else ""
        return Contact(name=child.text, email=email)

    @staticmethod
    def _parse_license(child: ET.Element, name: str) -> License:
        file = child.attrib["file"] if "file" in child.attrib else ""
        license_id = child.text
        lm = get_license_manager()
        normed_license = lm.normalize(license_id)
        if not normed_license:
            print(f"Unrecognized SPDX license ID specified for addon '{name}': '{license_id}'")
        elif normed_license != license_id:
            print(
                f"Unrecognized license string '{license_id}' normalized to SPDX license ID '{normed_license}' for addon '{name}'"
            )
            license_id = normed_license
        return License(name=license_id, file=file)

    @staticmethod
    def _parse_url(child: ET.Element) -> Url:
        url_type = UrlType.website
        branch = ""
        if "type" in child.attrib and child.attrib["type"] in UrlType.__dict__:
            url_type = UrlType[child.attrib["type"]]
            if url_type == UrlType.repository:
                branch = child.attrib["branch"] if "branch" in child.attrib else ""
        return Url(location=child.text, type=url_type, branch=branch)

    @staticmethod
    def _parse_dependency(child: ET.Element) -> Dependency:
        v_lt = child.attrib["version_lt"] if "version_lt" in child.attrib else ""
        v_lte = child.attrib["version_lte"] if "version_lte" in child.attrib else ""
        v_eq = child.attrib["version_eq"] if "version_eq" in child.attrib else ""
        v_gte = child.attrib["version_gte"] if "version_gte" in child.attrib else ""
        v_gt = child.attrib["version_gt"] if "version_gt" in child.attrib else ""
        condition = child.attrib["condition"] if "condition" in child.attrib else ""
        optional = "optional" in child.attrib and child.attrib["optional"].lower() == "true"
        dependency_type = DependencyType.automatic
        if "type" in child.attrib and child.attrib["type"] in DependencyType.__dict__:
            dependency_type = DependencyType[child.attrib["type"]]
        return Dependency(
            child.text,
            version_lt=v_lt,
            version_lte=v_lte,
            version_eq=v_eq,
            version_gte=v_gte,
            version_gt=v_gt,
            condition=condition,
            optional=optional,
            dependency_type=dependency_type,
        )

    @staticmethod
    def _parse_content(namespace: str, metadata: Metadata, root: ET.Element):
        """Given a content node, loop over its children, and if they are a recognized
        element type, recurse into each one to parse it."""
        known_content_types = ["workbench", "macro", "preferencepack", "bundle", "other"]
        for child in root:
            content_type = child.tag[len(namespace) :]
            if content_type in known_content_types:
                if content_type not in metadata.content:
                    metadata.content[content_type] = []
                metadata.content[content_type].append(MetadataReader._create_node(namespace, child))

    @staticmethod
    def _create_node(namespace, child) -> Metadata:
        new_content_item = Metadata()
        for content_child in child:
            MetadataReader._parse_child_element(namespace, content_child, new_content_item)
        return new_content_item


class MetadataWriter:
    """Utility class for serializing a Metadata object into the package.xml standard
    XML file."""

    @staticmethod
    def write(metadata: Metadata, path: str):
        """Write the metadata to a file located at path. Overwrites the file if it
        exists. Raises OSError if writing fails."""
        tree = MetadataWriter._create_tree_from_metadata(metadata)
        tree.write(path)

    @staticmethod
    def _create_tree_from_metadata(metadata: Metadata) -> ET.ElementTree:
        """Create the XML ElementTree representation of the given Metadata object."""
        tree = ET.ElementTree()
        root = tree.getroot()
        root.attrib["xmlns"] = "https://wiki.freecad.org/Package_Metadata"
        for key, value in metadata.__dict__.items():
            if isinstance(value, str):
                node = ET.SubElement(root, key)
                node.text = value
            else:
                MetadataWriter._create_list_node(metadata, key, root)
        return tree

    @staticmethod
    def _create_list_node(metadata: Metadata, key: str, root: ET.Element):
        for item in metadata.__dict__[key]:
            node = ET.SubElement(root, key)
            if key in ["maintainer", "author"]:
                if item.email:
                    node.attrib["email"] = item.email
                node.text = item.name
            elif key == "license":
                if item.file:
                    node.attrib["file"] = item.file
                node.text = item.name
            elif key == "url":
                if item.branch:
                    node.attrib["branch"] = item.branch
                node.attrib["type"] = str(item.type)
                node.text = item.location
            elif key in ["depend", "conflict", "replace"]:
                for dep_key, dep_value in item.__dict__.items():
                    if isinstance(dep_value, str) and dep_value:
                        node.attrib[dep_key] = dep_value
                    elif isinstance(dep_value, bool):
                        node.attrib[dep_key] = "True" if dep_value else "False"
                    elif isinstance(dep_value, DependencyType):
                        node.attrib[dep_key] = str(dep_value)



# ***************************************************************************
# *   Copyright (c) 2020 Sudhanshu Dubey <sudhanshu.thethunder@gmail.com>   *
# *   Copyright (c) 2021 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import sys
import FreeCAD
from FreeCAD import Rotation
from FreeCAD import Vector
from FreeCAD import Units

import Fem
import ObjectsFem
import Part
import Sketcher

from . import manager
from .manager import get_meshname
from .manager import init_doc
from .meshes import generate_mesh


def get_information():
    return {
        "name": "Electrostatics Electricforce - Elmer NonGUI6",
        "meshtype": "solid",
        "meshelement": "Tet10",
        "constraints": ["electrostatic potential"],
        "solvers": ["calculix", "elmer"],
        "material": "fluid",
        "equations": ["electrostatic"],
    }


def get_explanation(header=""):
    return (
        header
        + """

To run the example from Python console use:
from femexamples.equation_electrostatics_electricforce_elmer_nongui6 import setup
setup()


See forum topic post:
https://forum.freecad.org/viewtopic.php?f=18&t=41488&start=40#p373292

Electrostatics equation in FreeCAD FEM-Elmer

"""
    )


def setup(doc=None, solvertype="elmer"):

    # init FreeCAD document
    if doc is None:
        doc = init_doc()

    # explanation object
    # just keep the following line and change text string in get_explanation method
    manager.add_explanation_obj(doc, get_explanation(manager.get_header(get_information())))

    # geometric objects
    # name is important because the other method in this module use obj name
    geom_obj = doc.addObject("PartDesign::Body", "Body")
    base_sketch = geom_obj.newObject("Sketcher::SketchObject", "Base_Sketch")
    base_sketch.AttachmentSupport = (doc.getObject("XY_Plane"), [""])
    base_sketch.MapMode = "FlatFace"
    base_geoList = [
        Part.LineSegment(Vector(0.000000, 0.000000, 0), Vector(57.407921, 0.000000, 0)),
        Part.LineSegment(Vector(57.407921, 0.000000, 0), Vector(57.407921, 35.205284, 0)),
        Part.LineSegment(Vector(57.407921, 35.205284, 0), Vector(0.000000, 35.205284, 0)),
        Part.LineSegment(Vector(0.000000, 35.205284, 0), Vector(0.000000, 0.000000, 0)),
    ]
    base_sketch.addGeometry(base_geoList, False)
    base_conList = [
        Sketcher.Constraint("Coincident", 0, 2, 1, 1),
        Sketcher.Constraint("Coincident", 1, 2, 2, 1),
        Sketcher.Constraint("Coincident", 2, 2, 3, 1),
        Sketcher.Constraint("Coincident", 3, 2, 0, 1),
        Sketcher.Constraint("Horizontal", 0),
        Sketcher.Constraint("Horizontal", 2),
        Sketcher.Constraint("Vertical", 1),
        Sketcher.Constraint("Vertical", 3),
        Sketcher.Constraint("Coincident", 0, 1, -1, 1),
        Sketcher.Constraint("DistanceY", 1, 1, 1, 2, 35.205284),
        Sketcher.Constraint("DistanceX", 0, 1, 0, 2, 57.407921),
    ]
    base_sketch.addConstraint(base_conList)
    base_sketch.setDatum(9, Units.Quantity("5000.000000 mm"))
    base_sketch.setDatum(10, Units.Quantity("5000.000000 mm"))
    base_sketch.ViewObject.Visibility = False

    pad = geom_obj.newObject("PartDesign::Pad", "Pad")
    pad.Profile = base_sketch
    pad.Length = 7500.0
    pad.Length2 = 1000.0

    upper_sketch = geom_obj.newObject("Sketcher::SketchObject", "Upper_Sketch")
    upper_sketch.AttachmentSupport = None
    upper_sketch.MapMode = "Deactivated"
    upper_sketch.Placement = FreeCAD.Placement(Vector(0, 0, 1000), Rotation(Vector(0, 0, 1), 0))
    upper_geoList = [
        Part.LineSegment(Vector(25.560951, 4958.778320, 0), Vector(5068.406250, 4958.778320, 0)),
        Part.LineSegment(Vector(5068.406250, 4958.778320, 0), Vector(5037.082520, -21.422216, 0)),
        Part.LineSegment(Vector(5037.082520, 0.000000, 0), Vector(1309.763672, -21.422216, 0)),
        Part.LineSegment(Vector(1309.763672, 0.000000, 0), Vector(1372.406982, 1544.678467, 0)),
        Part.LineSegment(Vector(1372.406982, 1544.678467, 0), Vector(-37.083382, 1544.678467, 0)),
        Part.LineSegment(Vector(0.000000, 1544.678467, 0), Vector(25.560951, 4958.778320, 0)),
    ]
    upper_sketch.addGeometry(upper_geoList, False)
    upper_conList = [
        Sketcher.Constraint("Horizontal", 0),
        Sketcher.Constraint("Coincident", 1, 1, 0, 2),
        Sketcher.Constraint("PointOnObject", 1, 2, -1),
        Sketcher.Constraint("Vertical", 1),
        Sketcher.Constraint("Coincident", 2, 1, 1, 2),
        Sketcher.Constraint("PointOnObject", 2, 2, -1),
        Sketcher.Constraint("Coincident", 3, 1, 2, 2),
        Sketcher.Constraint("Coincident", 4, 1, 3, 2),
        Sketcher.Constraint("PointOnObject", 4, 2, -2),
        Sketcher.Constraint("Horizontal", 4),
        Sketcher.Constraint("Coincident", 5, 1, 4, 2),
        Sketcher.Constraint("Coincident", 5, 2, 0, 1),
        Sketcher.Constraint("Vertical", 5),
        Sketcher.Constraint("Vertical", 3),
        Sketcher.Constraint("DistanceX", 0, 1, 0, 2, 5037.082520),
        Sketcher.Constraint("DistanceY", 1, 2, 1, 1, 4958.778320),
        Sketcher.Constraint("DistanceY", 3, 1, 3, 2, 1544.678467),
        Sketcher.Constraint("DistanceX", 4, 2, 4, 1, 1309.763672),
    ]
    upper_sketch.addConstraint(upper_conList)
    upper_sketch.setDatum(14, Units.Quantity("5000.000000 mm"))
    upper_sketch.setDatum(15, Units.Quantity("5000.000000 mm"))
    upper_sketch.setDatum(16, Units.Quantity("1500.000000 mm"))
    upper_sketch.setDatum(17, Units.Quantity("1500.000000 mm"))
    upper_sketch.ViewObject.Visibility = False

    pocket = geom_obj.newObject("PartDesign::Pocket", "Pocket")
    pocket.Profile = upper_sketch
    pocket.Length = 1500.0
    pocket.Length2 = 100.0
    pocket.Reversed = 1
    doc.recompute()

    if FreeCAD.GuiUp:
        geom_obj.ViewObject.Document.activeView().viewAxonometric()
        geom_obj.ViewObject.Document.activeView().fitAll()

    # analysis
    analysis = ObjectsFem.makeAnalysis(doc, "Analysis")
    if FreeCAD.GuiUp:
        import FemGui

        FemGui.setActiveAnalysis(analysis)

    # solver
    if solvertype == "elmer":
        solver_obj = ObjectsFem.makeSolverElmer(doc, "SolverElmer")
        ObjectsFem.makeEquationElectrostatic(doc, solver_obj)
        ObjectsFem.makeEquationElectricforce(doc, solver_obj)
    elif solvertype == "calculix":
        solver_obj = ObjectsFem.makeSolverCalculiX(doc, "SolverCalculiX")
        solver_obj.AnalysisType = "electromagnetic"
    else:
        FreeCAD.Console.PrintWarning(
            "Unknown or unsupported solver type: {}. "
            "No solver object was created.\n".format(solvertype)
        )
    analysis.addObject(solver_obj)

    # material
    material_obj = ObjectsFem.makeMaterialFluid(doc, "Air")
    mat = material_obj.Material
    mat["Name"] = "Air"
    mat["Density"] = "1.204 kg/m^3"
    mat["ThermalConductivity"] = "0.02587 W/m/K"
    mat["ThermalExpansionCoefficient"] = "3.43e-3 1/K"
    mat["SpecificHeat"] = "1.01 kJ/kg/K"
    mat["ElectricalConductivity"] = "1e-12 S/m"
    mat["RelativePermeability"] = "1.0"
    mat["RelativePermittivity"] = "1.00059"
    material_obj.Material = mat
    material_obj.References = [(geom_obj, "Solid1")]
    analysis.addObject(material_obj)

    # constraint potential 0V
    name_pot1 = "ElectrostaticPotential1"
    con_elect_pot1 = ObjectsFem.makeConstraintElectrostaticPotential(doc, name_pot1)
    con_elect_pot1.References = [(geom_obj, "Face2")]
    con_elect_pot1.Potential = "0 V"
    con_elect_pot1.CapacitanceBody = 1
    con_elect_pot1.CapacitanceBodyEnabled = True
    con_elect_pot1.PotentialEnabled = True
    analysis.addObject(con_elect_pot1)

    # constraint potential 1V
    name_pot2 = "ElectrostaticPotential2"
    con_elect_pot2 = ObjectsFem.makeConstraintElectrostaticPotential(doc, name_pot2)
    con_elect_pot2.References = [(geom_obj, "Face4")]
    con_elect_pot2.Potential = "1 V"
    con_elect_pot2.CapacitanceBody = 2
    con_elect_pot2.CapacitanceBodyEnabled = True
    con_elect_pot2.PotentialEnabled = True
    con_elect_pot2.ElectricForcecalculation = True
    analysis.addObject(con_elect_pot2)

    # mesh
    femmesh_obj = analysis.addObject(ObjectsFem.makeMeshGmsh(doc, get_meshname()))[0]
    femmesh_obj.Shape = geom_obj
    femmesh_obj.SecondOrderLinear = False
    femmesh_obj.CharacteristicLengthMax = "500 mm"
    femmesh_obj.ViewObject.Visibility = False

    # mesh_region
    mesh_region = ObjectsFem.makeMeshRegion(doc, femmesh_obj, name="MeshRegion")
    mesh_region.CharacteristicLength = "250 mm"
    mesh_region.References = [
        (geom_obj, "Face4"),
        (geom_obj, "Face5"),
        (geom_obj, "Face6"),
        (geom_obj, "Face11"),
    ]
    mesh_region.ViewObject.Visibility = False

    # generate the mesh
    generate_mesh.mesh_from_mesher(femmesh_obj, "gmsh")

    doc.recompute()
    return doc



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2018 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************


"""The BIM Box command"""

import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP
translate = FreeCAD.Qt.translate


class BIM_Box:
    def GetResources(self):
        return {
            "Pixmap": "BIM_Box",
            "MenuText": QT_TRANSLATE_NOOP("BIM_Box", "Box"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "BIM_Box", "Graphically creates a generic box in the current document"
            ),
        }

    def IsActive(self):
        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):
        import WorkingPlane
        import draftguitools.gui_trackers as DraftTrackers

        FreeCAD.activeDraftCommand = self  # register as a Draft command for auto grid on/off
        self.doc = FreeCAD.ActiveDocument
        self.wp = WorkingPlane.get_working_plane()
        # here we will store our points
        self.points = []
        # we build a special cube tracker which is a list of 4 rectangle trackers
        self.cubetracker = []
        self.LengthValue = 0
        self.WidthValue = 0
        self.HeightValue = 0
        self.currentpoint = None
        for i in range(4):
            self.cubetracker.append(DraftTrackers.rectangleTracker())
        if hasattr(FreeCADGui, "Snapper"):
            FreeCADGui.Snapper.getPoint(
                callback=self.PointCallback,
                movecallback=self.MoveCallback,
                extradlg=self.taskbox(),
            )

    def MoveCallback(self, point, snapinfo):
        import DraftGeomUtils
        import DraftVecUtils

        self.currentpoint = point
        if len(self.points) == 1:
            # we have the base point already
            self.Length.setText(
                FreeCAD.Units.Quantity(
                    self.points[-1].sub(point).Length, FreeCAD.Units.Length
                ).UserString
            )
            self.Length.setFocus()
            self.Length.setSelection(0, FreeCADGui.draftToolBar.number_length(self.Length.text()))
        elif len(self.points) == 2:
            # now we already have our base line, we update the 1st rectangle
            p = point
            v1 = point.sub(self.points[1])
            v4 = v1.cross(self.points[1].sub(self.points[0]))
            if v4 and v4.Length:
                n = (self.points[1].sub(self.points[0])).cross(v4)
                if n and n.Length:
                    n = DraftVecUtils.project(v1, n)
                    p = self.points[1].add(n)
            self.cubetracker[0].p3(p)
            self.Width.setText(
                FreeCAD.Units.Quantity(
                    self.cubetracker[0].getSize()[1], FreeCAD.Units.Length
                ).UserString
            )
            self.Width.setFocus()
            self.Width.setSelection(0, FreeCADGui.draftToolBar.number_length(self.Width.text()))
        elif len(self.points) == 3:
            h = DraftGeomUtils.distance_to_plane(point, self.cubetracker[0].p3(), self.normal)
            w = self.normal * h
            # then we update all rectangles
            self.cubetracker[1].p3((self.cubetracker[0].p2()).add(w))
            self.cubetracker[2].p3((self.cubetracker[0].p4()).add(w))
            self.cubetracker[3].p1((self.cubetracker[0].p1()).add(w))
            self.cubetracker[3].p3((self.cubetracker[0].p3()).add(w))
            self.Height.setText(FreeCAD.Units.Quantity(h, FreeCAD.Units.Length).UserString)
            self.Height.setFocus()
            self.Height.setSelection(0, FreeCADGui.draftToolBar.number_length(self.Height.text()))

    def PointCallback(self, point, snapinfo):
        if not point:
            # cancelled
            self._finish()
            return

        if len(self.points) == 0:
            # this is our first clicked point, nothing to do just yet
            self.points.append(point)
            FreeCADGui.Snapper.getPoint(
                last=point,
                callback=self.PointCallback,
                movecallback=self.MoveCallback,
                extradlg=self.taskbox(),
            )
        elif len(self.points) == 1:
            # this is our second point
            baseline = point.sub(self.points[0])
            self.points.append(point)
            self._setupForWidthInput(baseline)
        elif len(self.points) == 2:
            # this is our third point
            self.points.append(point)
            self._setupForHeightInput()
        elif len(self.points) == 3:
            # finally we have all our points. Let's create the actual cube
            self._makeBox()
            self._finish()

    def taskbox(self):
        "sets up a taskbox widget"

        from PySide import QtGui

        wid = QtGui.QWidget()
        ui = FreeCADGui.UiLoader()
        wid.setWindowTitle(translate("BIM", "Box dimensions"))
        grid = QtGui.QGridLayout(wid)

        label1 = QtGui.QLabel(translate("BIM", "Length"))
        self.Length = ui.createWidget("Gui::InputField")
        l = FreeCAD.Units.Quantity(self.LengthValue, FreeCAD.Units.Length)
        self.Length.setText(l.UserString)
        grid.addWidget(label1, 0, 0, 1, 1)
        grid.addWidget(self.Length, 0, 1, 1, 1)
        if self.LengthValue:
            self.Length.setEnabled(False)

        label2 = QtGui.QLabel(translate("BIM", "Width"))
        self.Width = ui.createWidget("Gui::InputField")
        l = FreeCAD.Units.Quantity(self.WidthValue, FreeCAD.Units.Length)
        self.Width.setText(l.UserString)
        grid.addWidget(label2, 1, 0, 1, 1)
        grid.addWidget(self.Width, 1, 1, 1, 1)
        if self.WidthValue or (not self.LengthValue):
            self.Width.setEnabled(False)

        label3 = QtGui.QLabel(translate("BIM", "Height"))
        self.Height = ui.createWidget("Gui::InputField")
        l = FreeCAD.Units.Quantity(self.HeightValue, FreeCAD.Units.Length)
        self.Height.setText(l.UserString)
        grid.addWidget(label3, 2, 0, 1, 1)
        grid.addWidget(self.Height, 2, 1, 1, 1)
        if not self.WidthValue:
            self.Height.setEnabled(False)

        self.Length.textEdited.connect(FreeCADGui.draftToolBar.checkSpecialChars)
        self.Width.textEdited.connect(FreeCADGui.draftToolBar.checkSpecialChars)
        self.Height.textEdited.connect(FreeCADGui.draftToolBar.checkSpecialChars)
        self.Length.valueChanged.connect(self.setLength)
        self.Width.valueChanged.connect(self.setWidth)
        self.Height.valueChanged.connect(self.setHeight)
        self.Length.returnPressed.connect(self.setLengthUI)
        self.Width.returnPressed.connect(self.setWidthUI)
        self.Height.returnPressed.connect(self.setHeightUI)

        return wid

    def setLength(self, d):
        self.LengthValue = d

    def setWidth(self, d):
        self.WidthValue = d

    def setHeight(self, d):
        self.HeightValue = d

    def setLengthUI(self):
        if (len(self.points) == 1) and self.currentpoint and self.LengthValue:
            baseline = self.currentpoint.sub(self.points[0])
            baseline.normalize()
            baseline.multiply(self.LengthValue)
            p2 = self.points[0].add(baseline)
            self.points.append(p2)
            self._setupForWidthInput(baseline)

    def setWidthUI(self):
        if (len(self.points) == 2) and self.currentpoint and self.WidthValue:
            self.normal = self.cubetracker[0].getNormal()
            if self.normal:
                n = (self.points[1].sub(self.points[0])).cross(self.normal)
                if n and n.Length:
                    n.normalize()
                    n.multiply(self.WidthValue)
                    p2 = self.points[1].add(n)
            self.cubetracker[0].p3(p2)
            self.points.append(p2)
            self._setupForHeightInput()

    def setHeightUI(self):
        if (len(self.points) == 3) and self.HeightValue:
            self._makeBox()
            self._finish()

    def _setupForWidthInput(self, baseline):
        # we turn on only one of the rectangles
        self.cubetracker[0].setPlane(baseline)
        self.cubetracker[0].p1(self.points[0])
        self.cubetracker[0].on()
        FreeCADGui.Snapper.getPoint(
            last=self.points[-1],
            callback=self.PointCallback,
            movecallback=self.MoveCallback,
            extradlg=self.taskbox(),
        )

    def _setupForHeightInput(self):
        # we can get the cubes Z axis from our first rectangle
        self.normal = self.cubetracker[0].getNormal()
        # we can therefore define the (u,v) planes of all rectangles
        u = self.cubetracker[0].u
        v = self.cubetracker[0].v
        self.cubetracker[1].setPlane(u, self.normal)
        self.cubetracker[2].setPlane(u, self.normal)
        self.cubetracker[3].setPlane(u, v)
        # and the origin points of the vertical rectangles
        self.cubetracker[1].p1(self.cubetracker[0].p1())
        self.cubetracker[2].p1(self.cubetracker[0].p3())
        # finally we turn all rectangles on
        for r in self.cubetracker:
            r.on()
        point = self.cubetracker[0].p3()
        axis = self.cubetracker[0].p2().sub(self.cubetracker[0].p3())
        self.wp._save()
        self.wp.align_to_point_and_axis(point, axis, upvec=self.normal, _hist_add=False)
        FreeCADGui.Snapper.getPoint(
            last=point,
            callback=self.PointCallback,
            movecallback=self.MoveCallback,
            extradlg=self.taskbox(),
        )

    def _makeBox(self):
        import DraftGeomUtils

        p1 = self.cubetracker[0].p1()
        p2 = self.cubetracker[0].p2()
        p3 = self.cubetracker[0].p4()
        pla = DraftGeomUtils.placement_from_points(p1, p2, p3)
        if self.normal.isEqual(pla.Rotation.multVec(FreeCAD.Vector(0, 0, 1)), 1e-6):
            if self.HeightValue < 0.0:
                pla = DraftGeomUtils.placement_from_points(p1, p3, p2)
                self.LengthValue, self.WidthValue = self.WidthValue, self.LengthValue
        else:
            if self.HeightValue > 0.0:
                pla = DraftGeomUtils.placement_from_points(p1, p3, p2)
                self.LengthValue, self.WidthValue = self.WidthValue, self.LengthValue
        self.doc.openTransaction(translate("Arch", "Create Box"))
        cube = self.doc.addObject("Part::Box", "Box")
        cube.Placement = pla
        cube.Length = self.LengthValue
        cube.Width = self.WidthValue
        cube.Height = abs(self.HeightValue)
        self.doc.commitTransaction()
        self.doc.recompute()

    def _finish(self):
        self.wp._restore()
        FreeCAD.activeDraftCommand = None
        FreeCADGui.Snapper.off()
        for c in self.cubetracker:
            c.finalize()


FreeCADGui.addCommand("BIM_Box", BIM_Box())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2023 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""This NativeIFC module handles the retrieval and display of geometry compositions of objects"""

import FreeCAD

TAB = 2


def get_geometry_tree(element, prefix=""):
    """Returns a list of elements representing an object's representation"""

    result = [prefix + str(element)]
    prefix += TAB * " "
    if getattr(element, "Representation", None):
        reps = element.Representation
        result.append(prefix + str(reps))
        prefix += prefix
        for rep in reps.Representations:
            result.extend(get_geometry_tree(rep, prefix))
    elif getattr(element, "Items", None):
        for it in element.Items:
            result.extend(get_geometry_tree(it, prefix))
    elif element.is_a("IfcPolyline"):
        for p in element.Points:
            result.append(prefix + str(p))
    elif element.is_a("IfcExtrudedAreaSolid"):
        result.append(prefix + str(element.ExtrudedDirection))
        result.extend(get_geometry_tree(element.SweptArea, prefix))
    elif element.is_a("IfcArbitraryClosedProfileDef"):
        result.extend(get_geometry_tree(element.OuterCurve, prefix))
    elif element.is_a("IfcArbitraryProfileDefWithVoids"):
        result.extend(get_geometry_tree(element.OuterCurve, prefix))
        for inn in element.InnerCurves:
            result.extend(get_geometry_tree(inn, prefix))
    elif element.is_a("IfcMappedItem"):
        result.extend(get_geometry_tree(element.MappingSource[1], prefix))
    elif element.is_a("IfcBooleanClippingResult"):
        result.extend(get_geometry_tree(element.FirstOperand, prefix))
        result.extend(get_geometry_tree(element.SecondOperand, prefix))
    elif element.is_a("IfcBooleanResult"):
        result.extend(get_geometry_tree(element.FirstOperand, prefix))
        result.extend(get_geometry_tree(element.SecondOperand, prefix))
    elif element.is_a("IfcHalfSpaceSolid"):
        result.extend(get_geometry_tree(element.BaseSurface, prefix))
    return result


def print_geometry_tree(element):
    """Same as get_geometry_tree but printed"""

    for line in get_geometry_tree(element):
        print(line)


def show_geometry_tree(element):
    """Same as get_geometry_tree but in a Qt dialog"""

    import Arch_rc
    import FreeCADGui  # lazy import
    from . import ifc_tools
    from PySide import QtWidgets

    if isinstance(element, FreeCAD.DocumentObject):
        element = ifc_tools.get_ifc_element(element)
        if not element:
            return
    dlg = FreeCADGui.PySideUic.loadUi(":/ui/dialogTree.ui")
    tops = {}
    for line in get_geometry_tree(element):
        psize = (len(line) - len(line.lstrip())) / TAB
        wline = QtWidgets.QTreeWidgetItem([line.lstrip()])
        if not psize:
            dlg.geomtree.addTopLevelItem(wline)
        else:
            parent = tops[psize - 1]
            parent.addChild(wline)
        tops[psize] = wline
    dlg.geomtree.expandAll()
    dlg.setWindowTitle(dlg.windowTitle() + " " + element.Name)
    dlg.geomtree.currentItemChanged.connect(show_properties)
    dlg.proptree.itemChanged.connect(edit_property)
    result = dlg.exec_()
    if result:
        props = [r.split("::") for r in dlg.proptree.toolTip().split(";;;;") if r]
        if not props:
            return
        obj = ifc_tools.get_object(element)
        ifcfile = ifc_tools.get_ifcfile(obj)
        modified = False
        for prop in props:
            elt = ifcfile[int(prop[0])]
            attrib = prop[1]
            if attrib not in dir(elt):
                print("DEBUG: Unknown attribute:", attrib)
                continue
            value = prop[3]
            if isfloat(getattr(elt, attrib)):
                try:
                    value = float(value)
                except:
                    print("DEBUG: wrong value for", attrib, ":", value)
                    continue
            ifc_tools.set_attribute(ifcfile, elt, attrib, value)
            modified = True
        if modified:
            obj.touch()
            proj = ifc_tools.get_project(obj)
            proj.Modified = True
            obj.Document.recompute()


def isfloat(s):
    """Tells if the given string is a number"""

    try:
        float(s)
    except ValueError:
        return False
    print(s)
    return True


def show_properties(current, previous):
    """Displays object properties"""

    import FreeCADGui
    from . import ifc_tools  # lazy loading
    from PySide import QtCore, QtWidgets

    ifcid = int(current.text(0).split("=", 1)[0].strip(" ").strip("#"))
    sel = FreeCADGui.Selection.getSelection()
    if len(sel) != 1:
        return
    obj = sel[0]
    ifcfile = ifc_tools.get_ifcfile(obj)
    if not ifcfile:
        return
    elt = ifcfile[ifcid]
    box = current.treeWidget().parentWidget().widget(1)
    box.setTitle("#" + str(ifcid) + ": " + elt.is_a())
    props = [p for p in dir(elt) if p[0].isupper()]
    # allow only numeric values?
    # props = [p for p in  props if isfloat(str(getattr(elt,p)))]
    props = [p for p in props if not str(getattr(elt, p)).startswith("#")]
    props = [p for p in props if not str(getattr(elt, p)).startswith("(")]
    props = [p for p in props if p not in ["Position", "LayerAssignments", "StyledByItem"]]
    proptree = box.children()[0].itemAt(0).widget()
    proptree.clear()
    proptree.setHorizontalHeaderLabels(["Property", "Value"])
    proptree.setRowCount(len(props))
    for i, prop in enumerate(props):
        value = str(getattr(elt, prop))
        if value == "None":
            value = ""
        r1 = QtWidgets.QTableWidgetItem(prop)
        r1.setFlags(QtCore.Qt.ItemIsSelectable)
        proptree.setItem(i, 0, r1)
        r2 = QtWidgets.QTableWidgetItem(value)
        r2.setToolTip(value)
        if value.startswith("#") or value.startswith("(") or prop == "GlobalId":
            r2.setFlags(QtCore.Qt.ItemIsSelectable)
        proptree.setItem(i, 1, r2)
    if "Position" in dir(elt):
        i = proptree.rowCount()
        proptree.setRowCount(i + 6)
        position = FreeCAD.Vector(elt.Position.Location.Coordinates)
        axis = FreeCAD.Vector(elt.Position.Axis.DirectionRatios)
        xref = FreeCAD.Vector(elt.Position.RefDirection.DirectionRatios)
        rotation = FreeCAD.Rotation(axis, xref, FreeCAD.Vector(), "ZXY").toEulerAngles("XYZ")
        rotation = FreeCAD.Vector(rotation)
        for c in ["x", "y", "z"]:
            r1 = QtWidgets.QTableWidgetItem("Position " + c.upper())
            r1.setFlags(QtCore.Qt.ItemIsSelectable)
            proptree.setItem(i, 0, r1)
            v = str(getattr(position, c))
            r2 = QtWidgets.QTableWidgetItem(v)
            r2.setToolTip(v)
            proptree.setItem(i, 1, r2)
            i += 1
        for c in ["x", "y", "z"]:
            r1 = QtWidgets.QTableWidgetItem("Rotation " + c.upper())
            r1.setFlags(QtCore.Qt.ItemIsSelectable)
            proptree.setItem(i, 0, r1)
            v = str(getattr(rotation, c))
            r2 = QtWidgets.QTableWidgetItem(v)
            r2.setToolTip(v)
            proptree.setItem(i, 1, r2)
            i += 1


def edit_property(item):
    """Edits the value of a property"""

    if item.toolTip() and (item.text() != item.toolTip()):
        old = item.toolTip()
        new = item.text()
        table = item.tableWidget()
        prop = table.item(item.row(), 0).text()
        dlg = table.parent().parent().parent()
        line = dlg.geomtree.currentItem().text(0)
        ifcid = line.split("=", 1)[0].strip(" ").strip("#")
        strid = ";;;;" + "::".join([str(i) for i in [ifcid, prop, old, new]])
        dlg.proptree.setToolTip(dlg.proptree.toolTip() + strid)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2021 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import Part
import Path
import Path.Base.FeedRate as PathFeedRate
import Path.Base.MachineState as PathMachineState
from Path.Tool.toolbit import ToolBit
import Path.Tool.Controller as PathToolController
import PathScripts.PathUtils as PathUtils

from CAMTests.PathTestUtils import PathTestBase


def createTool(name="t1", diameter=1.75):
    attrs = {
        "name": name or "t1",
        "shape": "endmill.fcstd",
        "parameter": {"Diameter": diameter},
        "attribute": {},
    }
    toolbit = ToolBit.from_dict(attrs)
    return toolbit.attach_to_doc(doc=FreeCAD.ActiveDocument)


class TestPathHelpers(PathTestBase):
    def setUp(self):
        self.doc = FreeCAD.newDocument("TestPathUtils")

        c1 = Path.Command("G0 Z10")
        c2 = Path.Command("G0 X20 Y10")
        c3 = Path.Command("G1 X20 Y10 Z5")
        c4 = Path.Command("G1 X20 Y20")

        self.commandlist = [c1, c2, c3, c4]

    def tearDown(self):
        FreeCAD.closeDocument("TestPathUtils")

    def test00(self):
        """Test that FeedRate Helper populates horiz and vert feed rate based on TC"""
        t = createTool("test", 5.0)
        tc = PathToolController.Create("TC0", t)
        tc.VertRapid = 5
        tc.HorizRapid = 10
        tc.VertFeed = 15
        tc.HorizFeed = 20

        resultlist = PathFeedRate.setFeedRate(self.commandlist, tc)
        print(resultlist)

        self.assertTrue(resultlist[0].Parameters["F"] == 5)
        self.assertTrue(resultlist[1].Parameters["F"] == 10)
        self.assertTrue(resultlist[2].Parameters["F"] == 15)
        self.assertTrue(resultlist[3].Parameters["F"] == 20)

    def test01(self):
        """Test that Machine State initializes and stores position correctly"""

        machine = PathMachineState.MachineState()
        state = machine.getState()
        self.assertTrue(state["X"] == 0)
        self.assertTrue(state["Y"] == 0)
        self.assertTrue(state["Z"] == 0)
        self.assertTrue(machine.WCS == "G54")

        for c in self.commandlist:
            result = machine.addCommand(c)

        state = machine.getState()
        self.assertTrue(state["X"] == 20)
        self.assertTrue(state["Y"] == 20)
        self.assertTrue(state["Z"] == 5)

        machine.addCommand(Path.Command("M3 S200"))
        self.assertTrue(machine.S == 200)
        self.assertTrue(machine.Spindle == "CW")

        machine.addCommand(Path.Command("M4 S200"))
        self.assertTrue(machine.Spindle == "CCW")

        machine.addCommand(Path.Command("M2"))
        self.assertTrue(machine.Spindle == "off")
        self.assertTrue(machine.S == 0)

        machine.addCommand(Path.Command("G57"))
        self.assertTrue(machine.WCS == "G57")

        machine.addCommand(Path.Command("M6 T5"))
        self.assertTrue(machine.T == 5)

        # Test that non-change commands return false
        result = machine.addCommand(Path.Command("G0 X20"))
        self.assertFalse(result)

        result = machine.addCommand(Path.Command("G0 X30"))
        self.assertTrue(result)

        # Test that Drilling moves are handled correctly
        result = machine.addCommand(Path.Command("G81 X50 Y50 Z0"))
        state = machine.getState()
        self.assertTrue(state["X"] == 50)
        self.assertTrue(state["Y"] == 50)
        self.assertTrue(state["Z"] == 5)

    def test02(self):
        """Test PathUtils filterarcs"""

        # filter a full circle
        c = Part.Circle()
        c.Radius = 5
        edge = c.toShape()

        results = PathUtils.filterArcs(edge)
        self.assertTrue(len(results) == 2)
        e1 = results[0]
        self.assertTrue(isinstance(e1.Curve, Part.Circle))
        self.assertTrue(Path.Geom.pointsCoincide(edge.Curve.Location, e1.Curve.Location))
        self.assertTrue(edge.Curve.Radius == e1.Curve.Radius)

        # filter a 180 degree arc
        results = PathUtils.filterArcs(e1)
        self.assertTrue(len(results) == 1)

        # Handle a straight segment
        v1 = FreeCAD.Vector(0, 0, 0)
        v2 = FreeCAD.Vector(10, 0, 0)
        l = Part.makeLine(v1, v2)
        results = PathUtils.filterArcs(l)
        self.assertTrue(len(results) == 0)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2023 <https://www.freecad.org>                          *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""Provides GUI tools to open hyperlinks to internal/external documents."""

## @package gui_hyperlink
# \ingroup draftguitools
# \brief Provides GUI tools to open hyperlinks to internal/external documents

## \addtogroup draftguitools
# @{

import FreeCAD
import os
import re
from draftutils.messages import _msg, _toolmsg

if FreeCAD.GuiUp:
    import FreeCADGui
    import Draft_rc
    from PySide.QtCore import QUrl
    from PySide.QtGui import QDesktopServices


def QT_TRANSLATE_NOOP(ctx, txt):
    return txt


translate = FreeCAD.Qt.translate

from PySide import QtWidgets

__title__ = "FreeCAD Draft Workbench GUI Tools - Hyperlinks tools"
__author__ = ""
__url__ = "https://www.freecad.org"


class Draft_Hyperlink:
    """The Draft_Hyperlink FreeCAD command definition."""

    def GetResources(self):
        d = {
            "Pixmap": "",
            "Accel": "",
            "MenuText": QT_TRANSLATE_NOOP("Draft_Hyperlink", "Open Links"),
            "ToolTip": QT_TRANSLATE_NOOP("Draft_Hyperlink", "Opens linked documents"),
        }
        return d

    def Activated(self):
        self.find_hyperlinks()

        ret = None
        if len(self.hyperlinks_list) > 1:
            m = QtWidgets.QMessageBox()
            m.setWindowTitle(translate("draft", "Opening Multiple Links"))
            m.setText(translate("draft", "Multiple links found"))
            m.setInformativeText(
                translate("draft", "This may lead to the opening of various windows")
            )
            m.setStandardButtons(m.Ok | m.Cancel)
            ret = m.exec_()

        if len(self.hyperlinks_list) == 1 or ret == m.Ok:
            for hyperlink in self.hyperlinks_list:
                self.open_hyperlink(hyperlink)

    def find_hyperlinks(self):
        self.hyperlinks_list = []

        for o in FreeCADGui.Selection.getCompleteSelection():
            if hasattr(o.Object, "Text"):

                for text in o.Object.Text:
                    hyperlinks = re.findall(
                        r"((\w:[\\/]|%\w+%|\\\\\w+|/\w+|\w{3,5}://)[\w\\/: ]+\.[\S]+)", text
                    )

                    for hyperlink in hyperlinks:
                        self.hyperlinks_list.append(hyperlink[0])

    def has_hyperlinks(self):
        self.find_hyperlinks()

        return len(self.hyperlinks_list) > 0

    def open_hyperlink(self, hyperlink):
        file_hyperlink = len(re.findall(r"^(\w:[\\/]|%\w+%|\\\\\w+|/\w+)", hyperlink)) > 0

        url = None
        if file_hyperlink:
            if not os.path.isfile(hyperlink):
                _msg(translate("draft", "File not found:") + " " + hyperlink)
                return
            url = QUrl.fromLocalFile(hyperlink)
        else:
            url = QUrl(hyperlink)

        _toolmsg(translate("draft", "Opening hyperlink") + " " + hyperlink)

        QDesktopServices.openUrl(
            url
        )  # ToDo: add management to open FCStd files in the current instance and to open web pages with the Web Workbench


FreeCADGui.addCommand("Draft_Hyperlink", Draft_Hyperlink())

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""Document observer to act on documents containing NativeIFC objects"""

import FreeCAD


params = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/NativeIFC")


def add_observer():
    """Adds this observer to the running FreeCAD instance"""

    FreeCAD.BIMobserver = ifc_observer()
    FreeCAD.addDocumentObserver(FreeCAD.BIMobserver)


def remove_observer():
    """Removes this observer if present"""

    if hasattr(FreeCAD, "BIMobserver"):
        FreeCAD.removeDocumentObserver(FreeCAD.BIMobserver)
        del FreeCAD.BIMobserver


class ifc_observer:
    """A general document observer that handles IFC objects"""

    def __init__(self):
        # if there is a document open when the observer starts,
        # check it
        if FreeCAD.ActiveDocument:
            self.slotActivateDocument(FreeCAD.ActiveDocument)

    def slotStartSaveDocument(self, doc, value):
        """Save all IFC documents in this doc"""

        from PySide import QtCore  # lazy loading

        self.docname = doc.Name
        # delay execution to not get caught under the wait sursor
        # that occurs when the saveAs file dialog is shown
        # TODO find a more solid way
        QtCore.QTimer.singleShot(100, self.save)

    def slotDeletedObject(self, obj):
        """Deletes the corresponding object in the IFC document"""

        from . import ifc_tools  # lazy loading

        proj = ifc_tools.get_project(obj)
        if not proj:
            return
        if not hasattr(obj, "Proxy"):
            return
        if getattr(obj.Proxy, "nodelete", False):
            return
        ifc_tools.remove_ifc_element(obj)

    def slotChangedDocument(self, doc, prop):
        """Watch document IFC properties"""

        # only look at locked IFC documents
        if "IfcFilePath" not in doc.PropertiesList:
            return

        from . import ifc_tools  # lazy import

        if prop == "Schema":
            schema = doc.Schema
            ifcfile = ifc_tools.get_ifcfile(doc)
            if ifcfile:
                if schema != ifcfile.wrapped_data.schema_name():
                    # TODO display warming
                    ifcfile, migration_table = ifc_tools.migrate_schema(ifcfile, schema)
                    doc.Proxy.ifcfile = ifcfile
                    # migrate children
                    for old_id, new_id in migration_table.items():
                        child = [o for o in doc.Objects if getattr(o, "StepId", None) == old_id]
                        if len(child) == 1:
                            child[0].StepId = new_id
        elif prop == "Label":
            ifcfile = ifc_tools.get_ifcfile(doc)
            project = ifc_tools.get_ifc_element(doc)
            ifc_tools.set_attribute(ifcfile, project, "Name", doc.Label)

    def slotCreatedObject(self, obj):
        """If this is an IFC document, turn the object into IFC"""

        doc = getattr(obj, "Document", None)
        if doc:
            if hasattr(doc, "IfcFilePath"):
                from PySide import QtCore  # lazy loading

                self.objname = obj.Name
                self.docname = obj.Document.Name
                # delaying to make sure all other properties are set
                QtCore.QTimer.singleShot(100, self.convert)

    def slotActivateDocument(self, doc):
        """Check if we need to lock"""

        from . import ifc_status

        ifc_status.on_activate()

    def slotRemoveDynamicProperty(self, obj, prop):

        from . import ifc_psets

        ifc_psets.remove_property(obj, prop)

    # implementation methods

    def fit_all(self):
        """Fits the view"""

        if FreeCAD.GuiUp:
            import FreeCADGui

            FreeCADGui.SendMsgToActiveView("ViewFit")

    def save(self):
        """Saves all IFC documents contained in self.docname Document"""

        if not hasattr(self, "docname"):
            return
        if self.docname not in FreeCAD.listDocuments():
            return
        doc = FreeCAD.getDocument(self.docname)
        del self.docname
        projects = []
        if hasattr(doc, "IfcFilePath") and hasattr(doc, "Modified"):
            if doc.Modified:
                projects.append(doc)
        else:
            for obj in doc.findObjects(Type="Part::FeaturePython"):
                if hasattr(obj, "IfcFilePath") and hasattr(obj, "Modified"):
                    if obj.Modified:
                        projects.append(obj)
        if projects:
            from . import ifc_tools  # lazy loading
            from . import ifc_viewproviders

            ask = params.GetBool("AskBeforeSaving", True)
            if ask and FreeCAD.GuiUp:
                import Arch_rc
                import FreeCADGui

                dlg = FreeCADGui.PySideUic.loadUi(":/ui/dialogExport.ui")
                result = dlg.exec_()
                if not result:
                    return
                ask = dlg.checkAskBeforeSaving.isChecked()
                params.SetBool("AskBeforeSaving", ask)

            for project in projects:
                if getattr(project.Proxy, "ifcfile", None):
                    if project.IfcFilePath:
                        ifc_tools.save(project)
                    else:
                        ifc_viewproviders.get_filepath(project)
                        ifc_tools.save(project)

    def convert(self):
        """Converts an object to IFC"""

        if not hasattr(self, "objname") or not hasattr(self, "docname"):
            return
        if self.docname not in FreeCAD.listDocuments():
            return
        doc = FreeCAD.getDocument(self.docname)
        if not doc:
            return
        obj = doc.getObject(self.objname)
        if not obj:
            return
        if "StepId" in obj.PropertiesList:
            return
        del self.docname
        del self.objname
        if (
            obj.isDerivedFrom("Part::Feature")
            or "IfcType" in obj.PropertiesList
            or "CreateSpreadsheet" in obj.PropertiesList
        ):
            FreeCAD.Console.PrintLog("Converting " + obj.Label + " to IFC\n")
            from . import ifc_geometry  # lazy loading
            from . import ifc_tools  # lazy loading

            newobj = ifc_tools.aggregate(obj, doc)
            ifc_geometry.add_geom_properties(newobj)
            doc.recompute()



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""Widget for displaying a list of ToolBits using TwoLineTableCell."""

import yaml
import Path
from typing import Callable, List
from PySide import QtGui, QtCore
from PySide.QtGui import QDrag
from PySide.QtCore import QMimeData
from ..models.base import ToolBit
from .tablecell import TwoLineTableCell, CompactTwoLineTableCell


# Role for storing the ToolBit URI string
ToolBitUriRole = QtCore.Qt.UserRole + 1
ToolBitUriListMimeType = "application/x-freecad-toolbit-uri-list-yaml"


class ToolBitListWidget(QtGui.QListWidget):
    """
    A QListWidget specialized for displaying ToolBit items using
    TwoLineTableCell widgets.
    """

    def __init__(self, parent=None, tool_no_factory: Callable | None = None):
        super().__init__(parent)
        self._tool_no_factory = tool_no_factory
        self.setAutoScroll(True)
        self.setSelectionMode(QtGui.QAbstractItemView.ExtendedSelection)

    def setDragEnabled(self, enabled: bool = True):
        """Enable or disable drag-and-drop support for toolbits."""
        super().setDragEnabled(enabled)

    def startDrag(self, supportedActions):
        """Initiate drag with selected toolbits serialized as mime data if drag is enabled."""
        Path.Log.debug("startDrag: Drag initiated.")
        selected_items = self.selectedItems()
        if not selected_items:
            Path.Log.debug("startDrag: No items selected for drag.")
            return

        uris = [item.data(ToolBitUriRole) for item in selected_items]
        if not uris:
            Path.Log.debug("startDrag: No valid URIs found for selected items.")
            return

        # Create clipboard data
        clipboard_data = {
            "toolbits": uris,
        }
        yaml_data = yaml.safe_dump(clipboard_data).encode("utf-8")

        # Set mime data
        mime_data = QMimeData()
        mime_data.setData(ToolBitUriListMimeType, yaml_data)

        # Start drag
        drag = QDrag(self)
        drag.setMimeData(mime_data)
        drag.exec_(QtCore.Qt.CopyAction | QtCore.Qt.MoveAction)
        Path.Log.debug("startDrag: Drag executed.")

    def _create_toolbit_item(self, toolbit: ToolBit, tool_no: int | None = None):
        """
        Creates a QListWidgetItem and populates it with ToolBit data.
        """
        # Use the factory function if provided, otherwise use the passed tool_no
        final_tool_no = None
        if self._tool_no_factory:
            final_tool_no = self._tool_no_factory(toolbit)
        elif tool_no is not None:
            final_tool_no = tool_no

        # Add item to this widget
        item = QtGui.QListWidgetItem(self)
        cell = TwoLineTableCell(self)

        # Populate the cell widget
        cell.set_tool_no(final_tool_no)
        cell.set_upper_text(toolbit.label)
        cell.set_lower_text(toolbit.summary)
        cell.set_icon_from_shape(toolbit._tool_bit_shape)

        # Set the custom widget for the list item
        item.setSizeHint(cell.sizeHint())
        self.setItemWidget(item, cell)

        # Store the ToolBit URI for later retrieval
        item.setData(ToolBitUriRole, str(toolbit.get_uri()))

        return item

    def add_toolbit(self, toolbit: ToolBit, tool_no: int | None = None):
        """
        Adds a ToolBit to the list.

        Args:
            toolbit (ToolBit): The ToolBit object to add.
            tool_no (int | None): The tool number associated with the ToolBit,
                                  or None if not applicable.
        """
        item = self._create_toolbit_item(toolbit, tool_no)
        self.addItem(item)

    def insert_toolbit(self, row: int, toolbit: ToolBit, tool_no: int | None = None):
        """
        Inserts a ToolBit to the list at the specified row.

        Args:
            row (int): The row index where the item should be inserted.
            toolbit (ToolBit): The ToolBit object to add.
            tool_no (int | None): The tool number associated with the ToolBit,
                                  or None if not applicable.
        """
        item = self._create_toolbit_item(toolbit, tool_no)
        self.insertItem(row, item)

    def clear_list(self):
        """Removes all items from the list."""
        self.clear()

    def apply_filter(self, search_text: str):
        """
        Filters the list items based on the search text.
        Items are hidden if they don't contain the text in their
        tool number, upper text, or lower text.
        Also applies highlighting to the visible matching text.
        """
        search_text_lower = search_text.lower()
        for i in range(self.count()):
            item = self.item(i)
            cell = self.itemWidget(item)

            if isinstance(cell, TwoLineTableCell):
                cell.highlight(search_text)  # Apply highlighting
                # Determine visibility based on content
                contains = cell.contains_text(search_text_lower)
                item.setHidden(not contains)
            else:
                # Fallback for items without the expected widget (shouldn't happen)
                item_text = item.text().lower()  # Basic text search
                item.setHidden(search_text_lower not in item_text)

    def count_visible_items(self) -> int:
        """
        Counts and returns the number of visible items in the list.
        """
        visible_count = 0
        for i in range(self.count()):
            item = self.item(i)
            if not item.isHidden():
                visible_count += 1
        return visible_count

    def get_selected_toolbit_uri(self) -> str | None:
        """
        Returns the URI string of the currently selected ToolBit item.
        Returns None if no item is selected.
        """
        currentItem = self.currentItem()
        if currentItem:
            return currentItem.data(ToolBitUriRole)
        return None

    def get_selected_toolbit_uris(self) -> List[str]:
        """
        Returns a list of URI strings for the currently selected ToolBit items.
        Returns an empty list if no item is selected.
        """
        selected_uris = []
        selected_items = self.selectedItems()
        for item in selected_items:
            uri = item.data(ToolBitUriRole)
            if uri:
                selected_uris.append(uri)
        return selected_uris


class CompactToolBitListWidget(ToolBitListWidget):
    """
    A QListWidget specialized for displaying ToolBit items using
    CompactTwoLineTableCell widgets.
    """

    def _create_toolbit_item(self, toolbit: ToolBit, tool_no: int | None = None):
        """
        Creates a QListWidgetItem and populates it with ToolBit data
        using CompactTwoLineTableCell.
        """
        # Use the factory function if provided, otherwise use the passed tool_no
        final_tool_no = None
        if self._tool_no_factory:
            final_tool_no = self._tool_no_factory(toolbit)
        elif tool_no is not None:
            final_tool_no = tool_no

        item = QtGui.QListWidgetItem(self)
        cell = CompactTwoLineTableCell(self)

        # Populate the cell widget
        cell.set_tool_no(final_tool_no)
        cell.set_upper_text(toolbit.label)
        cell.set_lower_text(toolbit.summary)
        cell.set_icon_from_shape(toolbit._tool_bit_shape)

        # Set the custom widget for the list item
        item.setSizeHint(cell.sizeHint())
        self.setItemWidget(item, cell)

        # Store the ToolBit URI for later retrieval
        item.setData(ToolBitUriRole, str(toolbit.get_uri()))

        return item



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides various functions to work with offsets."""
## @package offsets
# \ingroup draftgeoutils
# \brief Provides various functions to work with offsets.

import lazy_loader.lazy_loader as lz

import FreeCAD as App
import DraftVecUtils

from draftgeoutils.general import geomType, vec
from draftgeoutils.geometry import get_normal
from draftgeoutils.wires import isReallyClosed
from draftgeoutils.intersections import wiresIntersect, connect

# Delay import of module until first use because it is heavy
Part = lz.LazyLoader("Part", globals(), "Part")

## \addtogroup draftgeoutils
# @{


def pocket2d(shape, offset):
    """Return a list of wires obtained from offsetting wires from the shape.

    Return a list of wires obtained from offsetting the wires
    from the given shape by the given offset, and intersection if needed.
    """
    # find the outer wire
    length = 0
    outerWire = None
    innerWires = []
    for w in shape.Wires:
        if w.BoundBox.DiagonalLength > length:
            outerWire = w
            length = w.BoundBox.DiagonalLength

    if not outerWire:
        return []

    for w in shape.Wires:
        if w.hashCode() != outerWire.hashCode():
            innerWires.append(w)

    o = outerWire.makeOffset(-offset)

    if not o.Wires:
        return []

    offsetWires = o.Wires
    # print("base offset wires:", offsetWires)
    if not innerWires:
        return offsetWires

    for innerWire in innerWires:
        i = innerWire.makeOffset(offset)
        if len(innerWire.Edges) == 1:
            e = innerWire.Edges[0]
            if isinstance(e.Curve, Part.Circle):
                e = Part.makeCircle(e.Curve.Radius + offset, e.Curve.Center, e.Curve.Axis)
                i = Part.Wire(e)
        if i.Wires:
            # print("offsetting island ", innerWire, " : ", i.Wires)
            for w in i.Wires:
                added = False
                # print("checking wire ",w)
                k = list(range(len(offsetWires)))
                for j in k:
                    # print("checking against existing wire ", j)
                    ow = offsetWires[j]
                    if ow:
                        if wiresIntersect(w, ow):
                            # print("intersect")
                            f1 = Part.Face(ow)
                            f2 = Part.Face(w)
                            f3 = f1.cut(f2)
                            # print("made new wires: ", f3.Wires)
                            offsetWires[j] = f3.Wires[0]
                            if len(f3.Wires) > 1:
                                # print("adding more")
                                offsetWires.extend(f3.Wires[1:])
                            added = True
                        else:
                            a = w.BoundBox
                            b = ow.BoundBox
                            if (
                                (a.XMin <= b.XMin)
                                and (a.YMin <= b.YMin)
                                and (a.ZMin <= b.ZMin)
                                and (a.XMax >= b.XMax)
                                and (a.YMax >= b.YMax)
                                and (a.ZMax >= b.ZMax)
                            ):
                                # print("this wire is bigger than "
                                #       "the outer wire")
                                offsetWires[j] = None
                                added = True
                            # else:
                            #   print("doesn't intersect")
                if not added:
                    # print("doesn't intersect with any other")
                    offsetWires.append(w)
    offsetWires = [o for o in offsetWires if o is not None]

    return offsetWires


def offset(edge, vector, trim=False):
    """Return a copy of the edge at a certain vector offset.

    If the edge is an arc, the vector will be added at its first point
    and a complete circle will be returned.

    None if there is a problem.

    Parameters
    ----------
    edge: Part.Shape
        the edge to offset
    vector: Base::Vector3
        the vector by which the edge is to be offset
    trim: bool, optional
        If `edge` is an arc and `trim` is `True`, the resulting
        arc will be trimmed to the proper angle

    Returns
    -------
    Part.Shape
        The offset shape
    """
    if not isinstance(edge, Part.Shape) or not isinstance(vector, App.Vector):
        return None

    if geomType(edge) == "Line":
        v1 = App.Vector.add(edge.Vertexes[0].Point, vector)
        v2 = App.Vector.add(edge.Vertexes[-1].Point, vector)
        return Part.LineSegment(v1, v2).toShape()

    elif geomType(edge) == "Circle":
        rad = edge.Vertexes[0].Point.sub(edge.Curve.Center)
        curve = Part.Circle(edge.Curve)
        curve.Radius = App.Vector.add(rad, vector).Length
        if trim:
            return Part.ArcOfCircle(curve, edge.FirstParameter, edge.LastParameter).toShape()
    elif geomType(edge) == "Ellipse":
        rad = edge.Vertexes[0].Point.sub(edge.Curve.Center)
        curve = edge.Curve.copy()
        if vector.getAngle(rad) < 1:
            curve.MajorRadius = curve.MajorRadius + vector.Length
            curve.MinorRadius = curve.MinorRadius + vector.Length
        else:
            curve.MajorRadius = curve.MajorRadius - vector.Length
            curve.MinorRadius = curve.MinorRadius - vector.Length
        return curve.toShape()
    else:
        return None


def offsetWire(
    wire,
    dvec,
    bind=False,
    occ=False,
    widthList=None,
    offsetMode=None,
    alignList=[],
    normal=None,
    basewireOffset=0,
    wireNedge=False,
):
    # normal=None, basewireOffset=0):
    """Offset the wire along the given vector.

    Parameters
    ----------
    wire as a sorted list of edges (the list is used directly), or as a
    wire or a face (Draft Wire with MakeFace True or False supported).

    The vector will be applied at the first vertex of the wire. If bind
    is True (and the shape is open), the original wire and the offsetted one
    are bound by 2 edges, forming a face.

    If widthList is provided (values only, not lengths - i.e. no unit),
    each value will be used to offset each corresponding edge in the wire.

    The 1st value overrides 'dvec' for 1st segment of wire;
    if a value is zero, value of 'widthList[0]' will follow;
    if widthList[0]' == 0, but dvec still provided, dvec will be followed

    offsetMode="BasewireMode" or None

    If alignList is provided,
    each value will be used to offset each corresponding edge
    in the wire with corresponding index.

    'basewireOffset' corresponds to 'offset' in ArchWall which offset
    the basewire before creating the wall outline; or
    it can be a list, e.g. corresponds to 'overrideOffset' in ArchWall, so
    offset can be 'per segment' of wire, or 'per wall segment' in ArchWall

    OffsetWire() is now aware of width and align per edge
    Primarily for use with ArchWall based on Sketch object

    To Do
    -----
    `dvec` vector to offset is now derived (and can be ignored)
    in this function if widthList and alignList are provided
    - 'dvec' to be obsolete in future?
    """

    if isinstance(wire, list) and isinstance(wire[0], Part.Edge):
        edges = wire.copy()
        wire = Part.Wire(edges)
        closed = wire.isClosed()
    elif isinstance(wire, Part.Wire):
        # Draft_Offset can fail when directly offsetting a Sketch wire. We need
        # to sort the edges. And because Part.__sortEdges__() can remove edges,
        # we need to create a new wire as well.
        edges = Part.__sortEdges__(wire.Edges)
        wire = Part.Wire(edges)
        closed = wire.isClosed()
    elif isinstance(wire, Part.Face):
        # We also need to sort the edges of a face.
        edges = Part.__sortEdges__(wire.OuterWire.Edges)
        closed = True
    elif isinstance(wire, Part.Edge):
        edges = [wire]
        closed = wire.isClosed()
    else:
        print("Either Part.Wire or Part.Edges should be provided, " "returning None")
        return None

    # For sketch with a number of wires, getNormal() may result
    # in different direction for each wire.
    # The 'normal' parameter, if provided e.g. by ArchWall,
    # allows normal over different wires e.g. in a Sketch be consistent
    # (over different calls of this function)
    if normal:
        norm = normal
    else:
        norm = get_normal(wire)  # norm = Vector(0, 0, 1)
        # for backward compatibility with previous getNormal implementation
        if norm is None:
            norm = App.Vector(0, 0, 1)

    nedges = []
    if occ:
        length = abs(dvec.Length)
        if not length:
            return None

        if wire.Wires:
            wire = wire.Wires[0]
        else:
            wire = Part.Wire(edges)

        try:
            off = wire.makeOffset(length)
        except Part.OCCError:
            return None
        else:
            return off

    # vec of first edge depends on its geometry
    e = edges[0]

    # Make a copy of alignList - to avoid changes in this function
    # become starting input of next call of this function?
    # https://www.dataquest.io/blog/tutorial-functions-modify-lists-dictionaries-python/
    # alignListC = alignList.copy()  # Only Python 3
    alignListC = list(alignList)  # Python 2 and 3

    # Check the direction / offset of starting edge
    firstDir = None
    try:
        if alignListC[0] == "Left":
            firstDir = 1
            firstAlign = "Left"
        elif alignListC[0] == "Right":
            firstDir = -1
            firstAlign = "Right"
        elif alignListC[0] == "Center":
            firstDir = 1
            firstAlign = "Center"
    except IndexError:
        # Should no longer happen for ArchWall
        # as aligns are 'filled in' by ArchWall
        pass

    # If not provided by alignListC checked above, check the direction
    # of offset in dvec (not 'align').

    # TODO Should check if dvec is provided or not
    # ('legacy/backward-compatible' mode)
    if not firstDir:
        # need to test against Part.Circle, not Part.ArcOfCircle
        if isinstance(e.Curve, (Part.Circle, Part.Ellipse)):
            v0 = e.tangentAt(e.FirstParameter).cross(norm)
        else:
            v0 = vec(e).cross(norm)
        # check against dvec provided for the offset direction
        # would not know if dvec is vector of width (Left/Right Align)
        # or width/2 (Center Align)
        v0.normalize()
        v1 = App.Vector(dvec).normalize()
        if v0.isEqual(v1, 0.0001):
            # "Left Offset" (Left Align or 'left offset' in Centre Align)
            firstDir = 1
            firstAlign = "Left"
            alignListC.append("Left")
        elif v0.isEqual(v1.negative(), 0.0001):
            # "Right Offset" (Right Align or 'right offset' in Centre Align)
            firstDir = -1
            firstAlign = "Right"
            alignListC.append("Right")
        else:
            print(" something wrong with firstDir ")
            firstAlign = "Left"
            alignListC.append("Left")

    if not isinstance(basewireOffset, list):
        basewireOffset = [basewireOffset]

    for i in range(len(edges)):
        # make a copy so it do not reverse the self.baseWires edges
        # pointed to by _Wall.getExtrusionData()?
        curredge = Part.Shape(edges[i]).Edges[0]  # saved memory?

        # record first edge's Orientation, Dir, Align and set Delta
        if i == 0:
            # TODO Could be edge.Orientation in fact
            # "Forward" or "Reversed"
            firstOrientation = curredge.Vertexes[0].Orientation
            curOrientation = firstOrientation
            curDir = firstDir
            curAlign = firstAlign
            delta = dvec

        # record current edge's Orientation, and set Delta
        if i != 0:  # else:
            # TODO Should also calculate 1st edge direction above
            if isinstance(curredge.Curve, (Part.Circle, Part.Ellipse)):
                # Seems Circle/Ellipse (1 geometry) would be sorted in 1 list,
                # so i=0; should not happen here in fact (i != 0). 2024.8.25.
                delta = curredge.tangentAt(curredge.FirstParameter).cross(norm)
            else:
                delta = vec(curredge).cross(norm)
            # TODO Could be edge.Orientation in fact
            curOrientation = curredge.Vertexes[0].Orientation

        # Consider individual edge width
        if widthList:  # ArchWall should now always provide widthList
            try:
                if widthList[i] > 0:
                    delta = DraftVecUtils.scaleTo(delta, widthList[i])
                elif dvec:
                    delta = DraftVecUtils.scaleTo(delta, dvec.Length)
                else:
                    # just hardcoded default value as ArchWall would provide
                    # if dvec is not provided either
                    delta = DraftVecUtils.scaleTo(delta, 200)
            except Part.OCCError:
                if dvec:
                    delta = DraftVecUtils.scaleTo(delta, dvec.Length)
                else:
                    # just hardcoded default value as ArchWall would provide
                    # if dvec is not provided either
                    delta = DraftVecUtils.scaleTo(delta, 200)
        else:
            delta = DraftVecUtils.scaleTo(delta, dvec.Length)

        # Consider individual offset distance
        try:
            currOffset = basewireOffset[i]
        except:
            currOffset = basewireOffset[0]  # use first value

        # Consider individual edge Align direction
        # - ArchWall should now always provide alignList
        if i == 0:
            if alignListC[0] == "Center":
                delta = DraftVecUtils.scaleTo(delta, delta.Length / 2)
            # No need to do anything for 'Left' and 'Right' as original dvec
            # have set both the direction and amount of offset correct
            # elif alignListC[i] == 'Left':  #elif alignListC[i] == 'Right':
        if i != 0:
            try:
                if alignListC[i] == "Left":
                    curDir = 1
                    curAlign = "Left"
                elif alignListC[i] == "Right":
                    curDir = -1
                    curAlign = "Right"
                    delta = delta.negative()
                elif alignListC[i] == "Center":
                    curDir = 1
                    curAlign = "Center"
                    delta = DraftVecUtils.scaleTo(delta, delta.Length / 2)
            except IndexError:
                curDir = firstDir
                curAlign = firstAlign
                if firstAlign == "Right":
                    delta = delta.negative()
                elif firstAlign == "Center":
                    delta = DraftVecUtils.scaleTo(delta, delta.Length / 2)

        # Consider whether generating the 'offset wire' or the 'base wire'
        if offsetMode is None:
            # Consider if curOrientation and/or curDir match their
            # firstOrientation/firstDir - to determine whether
            # and how to offset the current edge

            # This is a xor
            if (curOrientation == firstOrientation) != (curDir == firstDir):
                if curAlign in ["Left", "Right"]:
                    # ArchWall has an Offset properties for user to offset
                    # the basewire before creating the base profile of wall
                    # (not applicable to 'Center' align)
                    if currOffset:
                        delta = DraftVecUtils.scaleTo(delta, currOffset)
                        nedge = offset(curredge, delta, trim=True)
                    else:
                        nedge = curredge
                elif curAlign == "Center":
                    delta = delta.negative()
                    nedge = offset(curredge, delta, trim=True)
            else:
                # if curAlign in ['Left', 'Right']:
                # elif curAlign == 'Center': # Both conditions same result.
                # ArchWall has an Offset properties for user to offset
                # the basewire before creating the base profile of wall
                # (not applicable to 'Center' align)
                if currOffset:
                    if curAlign in ["Left", "Right"]:
                        delta = DraftVecUtils.scaleTo(delta, delta.Length + currOffset)
                nedge = offset(curredge, delta, trim=True)

            # TODO arc always in counter-clockwise directinon
            # ... ( not necessarily 'reversed')
            if curOrientation == "Reversed":
                if not isinstance(curredge.Curve, (Part.Circle, Part.Ellipse)):
                    # assume straight line, reverse it
                    nedge = Part.Edge(nedge.Vertexes[1], nedge.Vertexes[0])
                elif nedge.isClosed():
                    pass
                else:
                    midParameter = (
                        nedge.FirstParameter + (nedge.LastParameter - nedge.FirstParameter) / 2
                    )
                    midOfArc = nedge.valueAt(midParameter)
                    nedge = Part.ArcOfCircle(
                        nedge.Vertexes[1].Point, midOfArc, nedge.Vertexes[0].Point
                    ).toShape()
                    # TODO any better solution than to calculate midpoint
                    # of arc to reverse?

        elif offsetMode in ["BasewireMode"]:
            if not (curOrientation == firstOrientation) != (curDir == firstDir):
                if curAlign in ["Left", "Right"]:
                    # ArchWall has an Offset properties for user to offset
                    # the basewire before creating the base profile of wall
                    # (not applicable to 'Center' align)
                    if currOffset:
                        delta = DraftVecUtils.scaleTo(delta, currOffset)
                        nedge = offset(curredge, delta, trim=True)
                    else:
                        nedge = curredge
                elif curAlign == "Center":
                    delta = delta.negative()
                    nedge = offset(curredge, delta, trim=True)
            else:
                if curAlign in ["Left", "Right"]:
                    # ArchWall has an Offset properties for user to offset
                    # the basewire before creating the base profile of wall
                    # (not applicable to 'Center' align)
                    if currOffset:
                        delta = DraftVecUtils.scaleTo(delta, delta.Length + currOffset)
                    nedge = offset(curredge, delta, trim=True)
                elif curAlign == "Center":
                    nedge = offset(curredge, delta, trim=True)
            if curOrientation == "Reversed":
                if not isinstance(curredge.Curve, (Part.Circle, Part.Ellipse)):
                    # assume straight line, reverse it
                    nedge = Part.Edge(nedge.Vertexes[1], nedge.Vertexes[0])
                elif nedge.isClosed():
                    pass
                else:
                    midParameter = (
                        nedge.FirstParameter + (nedge.LastParameter - nedge.FirstParameter) / 2
                    )
                    midOfArc = nedge.valueAt(midParameter)
                    nedge = Part.ArcOfCircle(
                        nedge.Vertexes[1].Point, midOfArc, nedge.Vertexes[0].Point
                    ).toShape()
                    # TODO any better solution than to calculate midpoint
                    # of arc to reverse?
        else:
            print(" something wrong ")
            return None
        if not nedge:
            return None

        # nedges is offset edges
        nedges.append(nedge)

    if len(edges) > 1:
        # TODO May phase out wire if bind() can do without it later and do with
        # only connectEdges so no need bind() to find 'touching edges' there
        wire, connectEdgesF, connectEdges = connect(nedges, closed, wireNedge=True)
    else:
        # TODO May phase out wire if bind() can do without it later and do with
        # only connectEdges so no need bind() to find 'touching edges' there
        wire = Part.Wire(nedges[0])
        connectEdgesF = connectEdges = nedges  # nedges[0]

    if bind and not closed:
        e1 = Part.LineSegment(edges[0].Vertexes[0].Point, wire[0].Vertexes[0].Point).toShape()
        e2 = Part.LineSegment(edges[-1].Vertexes[-1].Point, wire[-1].Vertexes[-1].Point).toShape()
        # nedges[-1].Vertexes[-1].Point).toShape()
        alledges = edges.extend(nedges)  # TODO nedges is a wire or are edges?
        alledges = alledges.extend([e1, e2])
        w = Part.Wire(alledges)
        return w
    else:
        if wireNedge:
            return (wire, connectEdgesF, connectEdges, nedges)
        else:
            return wire


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# MeshPart gui init module
# (c) 2003 Juergen Riegel
#
# Gathering all the information to start FreeCAD
# This is the second one of three init scripts, the third one
# runs when the gui is up

# ***************************************************************************
# *   Copyright (c) 2002 Juergen Riegel <juergen.riegel@web.de>             *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Lesser General Public License for more details.                   *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************/


class MeshPartWorkbench(Workbench):
    "MeshPart workbench object"

    Icon = """
            /* XPM */
            static const char *MeshPart_Box[]={
            "16 16 3 1",
            ". c None",
            "# c #000000",
            "a c #c6c642",
            "................",
            ".......#######..",
            "......#aaaaa##..",
            ".....#aaaaa###..",
            "....#aaaaa##a#..",
            "...#aaaaa##aa#..",
            "..#aaaaa##aaa#..",
            ".########aaaa#..",
            ".#aaaaa#aaaaa#..",
            ".#aaaaa#aaaa##..",
            ".#aaaaa#aaa##...",
            ".#aaaaa#aa##....",
            ".#aaaaa#a##... .",
            ".#aaaaa###......",
            ".########.......",
            "................"};
            """
    MenuText = "MeshPart"
    ToolTip = "MeshPart workbench"

    def Initialize(self):
        # load the module
        import MeshPartGui
        import MeshPart

    def GetClassName(self):
        return "MeshPartGui::Workbench"


# Gui.addWorkbench(MeshPartWorkbench())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2024 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************


from . import ifc_tools  # lazy import


def edit_classification(obj):
    """Edits the classification of this object"""

    element = ifc_tools.get_ifc_element(obj)
    ifcfile = ifc_tools.get_ifcfile(obj)
    if not element or not ifcfile:
        return
    # TODO: remove previous reference?
    # ifc_tools.api_run("classification.remove_reference",
    #                   ifcfile, reference=ref, products=[obj])
    classifications = ifcfile.by_type("IfcClassification")
    classification = getattr(obj, "Classification", "")
    if classification:
        cname, code = classification.split(" ", 1)
        cnames = [c.Name for c in classifications]
        if cname in cnames:
            system = classifications[cnames.index(cname)]
        else:
            system = ifc_tools.api_run(
                "classification.add_classification", ifcfile, classification=cname
            )
        for ref in getattr(system, "HasReferences", []):
            rname = ref.Name or ref.Identification
            if code == rname:
                return
            elif code.startswith(rname):
                if getattr(ref, "ClassificationRefForObjects", None):
                    rel = ref.ClassificationRefForObjects[0]
                    if not element in rel.RelatedObjects:
                        ifc_tools.edit_attribute(
                            rel, "RelatedObjects", rel.RelatedObjects + [element]
                        )
                else:
                    # we have a reference, but no classForObjects
                    # this is weird and shouldn't exist...
                    rel = ifcfile.createIfcRelAssociatesClassification(
                        ifc_tools.ifcopenshell.guid.new(),
                        history,
                        "FreeCADClassificationRel",
                        None,
                        [element],
                        ref,
                    )
        else:
            ifc_tools.api_run(
                "classification.add_reference",
                ifcfile,
                products=[element],
                classification=system,
                identification=code,
            )
    else:
        # classification property is empty
        for rel in getattr(element, "HasAssociations", []):
            if rel.is_a("IfcRelAssociatesClassification"):
                # removing existing classification if only user
                if len(rel.RelatedObjects) == 1 and rel.RelatedObjects[0] == element:
                    ifc_tools.api_run(
                        "classification.remove_reference",
                        ifcfile,
                        reference=rel.RelatingClassification,
                        products=[element],
                    )
            # TODO: Remove IfcClassification too?


def show_classification(obj):
    """Loads the classification of this object"""

    element = ifc_tools.get_ifc_element(obj)
    ifcfile = ifc_tools.get_ifcfile(obj)
    if not element or not ifcfile:
        return
    for system in ifcfile.by_type("IfcClassification"):
        for ref in getattr(system, "HasReferences", []):
            for rel in ref.ClassificationRefForObjects:
                if element in rel.RelatedObjects:
                    if not "Classification" in obj.PropertiesList:
                        obj.addProperty("App::PropertyString", "Classification", "IFC", locked=True)
                    sname = system.Name
                    cname = ref.Name or ref.Identification
                    obj.Classification = sname + " " + cname
                    break



# ***************************************************************************
# *   Copyright (c) 2021 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM calculix constraint force"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"


def get_analysis_types():
    return ["buckling", "static", "thermomech"]


def get_sets_name():
    return "constraints_force_node_loads"


def get_before_write_meshdata_constraint():
    return ""


def get_after_write_meshdata_constraint():
    return ""


def write_meshdata_constraint(f, femobj, force_obj, ccxwriter):

    # floats read from ccx should use {:.13G}, see comment in writer module

    if force_obj.EnableAmplitude:
        f.write(f"*CLOAD, AMPLITUDE={force_obj.Name}\n")
    else:
        f.write("*CLOAD\n")
    direction_vec = femobj["Object"].DirectionVector
    dir_zero_tol = 1e-15  # TODO: should this be more generally for more values?
    # be careful with raising the tolerance, a big load would have an impact
    # but compared to the real direction the impact would be small again
    for ref_shape in femobj["NodeLoadTable"]:
        f.write(f"** {ref_shape[0]}\n")
        for n in sorted(ref_shape[1]):
            node_load = ref_shape[1][n]
            # the loads in ref_shape[1][n] are without unit
            if abs(direction_vec.x) > dir_zero_tol:
                v1 = f"{(direction_vec.x * node_load).Value:.13G}"
                f.write(f"{n},1,{v1}\n")
            if abs(direction_vec.y) > dir_zero_tol:
                v2 = f"{(direction_vec.y * node_load).Value:.13G}"
                f.write(f"{n},2,{v2}\n")
            if abs(direction_vec.z) > dir_zero_tol:
                v3 = f"{(direction_vec.z * node_load).Value:.13G}"
                f.write(f"{n},3,{v3}\n")
        f.write("\n")
    f.write("\n")



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *   Copyright (c) 2020 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

from PySide.QtCore import QT_TRANSLATE_NOOP
import FreeCAD
import Path
import re

__title__ = "Generic property container to store some values."
__author__ = "sliptonic (Brad Collette)"
__url__ = "https://www.freecad.org"
__doc__ = "A generic container for typed properties in arbitrary categories."

if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


translate = FreeCAD.Qt.translate


SupportedPropertyType = {
    "Angle": "App::PropertyAngle",
    "Bool": "App::PropertyBool",
    "Distance": "App::PropertyDistance",
    "Enumeration": "App::PropertyEnumeration",
    "File": "App::PropertyFile",
    "Float": "App::PropertyFloat",
    "Integer": "App::PropertyInteger",
    "Length": "App::PropertyLength",
    "Percent": "App::PropertyPercent",
    "String": "App::PropertyString",
}


def getPropertyTypeName(o):
    for typ in SupportedPropertyType:
        if SupportedPropertyType[typ] == o:
            return typ
    raise IndexError()


class PropertyBag(object):
    """Property container object."""

    CustomPropertyGroups = "CustomPropertyGroups"
    CustomPropertyGroupDefault = "User"

    def __init__(self, obj):
        # Always add as enumeration
        if not hasattr(obj, self.CustomPropertyGroups):
            obj.addProperty(
                "App::PropertyEnumeration",
                self.CustomPropertyGroups,
                "Base",
                QT_TRANSLATE_NOOP("App::Property", "List of custom property groups"),
            )
        self.onDocumentRestored(obj)

    def dumps(self):
        return None

    def loads(self, state):
        return None

    def __sanitizePropertyName(self, name):
        if len(name) == 0:
            return
        clean = name[0]
        for i in range(1, len(name)):
            if name[i] == " ":
                clean += name[i + 1].upper()
                i += 1
            elif name[i - 1] != " ":
                clean += name[i]
        return clean

    def onDocumentRestored(self, obj):
        self.obj = obj
        cpg = getattr(obj, self.CustomPropertyGroups, None)
        # If it's a string list, convert to enum
        if isinstance(cpg, list):
            vals = cpg
            try:
                obj.removeProperty(self.CustomPropertyGroups)
            except Exception:
                # Removing the property may fail if it does not exist; safe to ignore in this context.
                pass
            obj.addProperty(
                "App::PropertyEnumeration",
                self.CustomPropertyGroups,
                "Base",
                QT_TRANSLATE_NOOP("App::Property", "List of custom property groups"),
            )
            if hasattr(obj, "setEnumerationsOfProperty"):
                obj.setEnumerationsOfProperty(self.CustomPropertyGroups, vals)
            else:
                # Fallback: set the property value directly (may not work in all FreeCAD versions)
                setattr(obj, self.CustomPropertyGroups, vals)
            if hasattr(obj, "setEditorMode"):
                obj.setEditorMode(self.CustomPropertyGroups, 2)  # hide
        elif hasattr(obj, "getEnumerationsOfProperty"):
            if hasattr(obj, "setEditorMode"):
                obj.setEditorMode(self.CustomPropertyGroups, 2)  # hide

    def getCustomProperties(self):
        """Return a list of all custom properties created in this container."""
        groups = []
        if hasattr(self.obj, "getEnumerationsOfProperty"):
            groups = list(self.obj.getEnumerationsOfProperty(self.CustomPropertyGroups))
        else:
            groups = list(getattr(self.obj, self.CustomPropertyGroups, []))
        return [p for p in self.obj.PropertiesList if self.obj.getGroupOfProperty(p) in groups]

    def addCustomProperty(self, propertyType, name, group=None, desc=None):
        """addCustomProperty(propertyType, name, group=None, desc=None) ... adds a custom property and tracks its group."""
        if desc is None:
            desc = ""
        if group is None:
            group = self.CustomPropertyGroupDefault

        # Always use enum for groups
        if hasattr(self.obj, "getEnumerationsOfProperty"):
            groups = list(self.obj.getEnumerationsOfProperty(self.CustomPropertyGroups))
        else:
            groups = list(getattr(self.obj, self.CustomPropertyGroups, []))

        name = self.__sanitizePropertyName(name)
        if not re.match("^[A-Za-z0-9_]*$", name):
            raise ValueError("Property Name can only contain letters and numbers")

        if group not in groups:
            groups.append(group)
            if hasattr(self.obj, "setEnumerationsOfProperty"):
                self.obj.setEnumerationsOfProperty(self.CustomPropertyGroups, groups)
            else:
                setattr(self.obj, self.CustomPropertyGroups, groups)
        self.obj.addProperty(propertyType, name, group, desc)
        return name

    def refreshCustomPropertyGroups(self):
        """refreshCustomPropertyGroups() ... removes empty property groups, should be called after deleting properties."""
        customGroups = []
        for p in self.obj.PropertiesList:
            group = self.obj.getGroupOfProperty(p)
            if hasattr(self.obj, "getEnumerationsOfProperty"):
                groups = list(self.obj.getEnumerationsOfProperty(self.CustomPropertyGroups))
            else:
                groups = list(getattr(self.obj, self.CustomPropertyGroups, []))
            if group in groups and group not in customGroups:
                customGroups.append(group)
        if hasattr(self.obj, "setEnumerationsOfProperty"):
            self.obj.setEnumerationsOfProperty(self.CustomPropertyGroups, customGroups)
        else:
            setattr(self.obj, self.CustomPropertyGroups, customGroups)


def Create(name="PropertyBag"):
    obj = FreeCAD.ActiveDocument.addObject("App::FeaturePython", name)
    obj.Proxy = PropertyBag(obj)
    return obj


def IsPropertyBag(obj):
    """Returns True if the supplied object is a property container (or its Proxy)."""

    if type(obj) == PropertyBag:
        return True
    if hasattr(obj, "Proxy"):
        return IsPropertyBag(obj.Proxy)
    return False



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
import abc
from typing import List
from ..uri import AssetUri


class AssetStore(abc.ABC):
    """
    Abstract base class for storing and retrieving asset data as raw bytes.

    Stores are responsible for handling the low-level interaction with a
    specific storage backend (e.g., local filesystem, HTTP server) based
    on the URI protocol.
    """

    def __init__(self, name: str, *args, **kwargs):
        self.name = name

    @abc.abstractmethod
    async def get(self, uri: AssetUri) -> bytes:
        """
        Retrieve the raw byte data for the asset at the given URI.

        Args:
            uri: The unique identifier for the asset.

        Returns:
            The raw byte data of the asset.

        Raises:
            FileNotFoundError: If the asset does not exist at the URI.
            # Other store-specific exceptions may be raised.
        """
        raise NotImplementedError

    async def exists(self, uri: AssetUri) -> bool:
        """
        Check if the asset exists at the given URI.

        Args:
            uri: The unique identifier for the asset.

        Returns:
            True if the asset exists, False otherwise.
        """
        try:
            await self.get(uri)
            return True
        except FileNotFoundError:
            return False

    @abc.abstractmethod
    async def delete(self, uri: AssetUri) -> None:
        """
        Delete the asset at the given URI.

        Args:
            uri: The unique identifier for the asset to delete.

        Raises:
            FileNotFoundError: If the asset does not exist at the URI.
            # Other store-specific exceptions may be raised.
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def create(self, asset_type: str, asset_id: str, data: bytes) -> AssetUri:
        """
        Create a new asset in the store with the given data.

        The store determines the final URI for the new asset. The
        `asset_type` can be used to influence the storage location
        or URI structure (e.g., as part of the path).

        Args:
            asset_type: The type of the asset (e.g., 'material',
                           'toolbitshape').
            asset_id: The unique identifier for the asset.
            data: The raw byte data of the asset to create.

        Returns:
            The URI of the newly created asset.

        Raises:
            # Store-specific exceptions may be raised (e.g., write errors).
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def update(self, uri: AssetUri, data: bytes) -> AssetUri:
        """
        Update the asset at the given URI with new data, creating a new version.

        Args:
            uri: The unique identifier of the asset to update.
            data: The new raw byte data for the asset.

        Raises:
            FileNotFoundError: If the asset does not exist at the URI.
            # Other store-specific exceptions may be raised (e.g., write errors).
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def list_assets(
        self, asset_type: str | None = None, limit: int | None = None, offset: int | None = None
    ) -> List[AssetUri]:
        """
        List assets in the store, optionally filtered by asset type and
        with pagination. For versioned stores, this lists the latest
        version of each asset.

        Args:
            asset_type: Optional filter for asset type.
            limit: Maximum number of assets to return.
            offset: Number of assets to skip from the beginning.

        Returns:
            A list of URIs for the assets.
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def count_assets(self, asset_type: str | None = None) -> int:
        """
        Counts assets in the store, optionally filtered by asset type.

        Args:
            asset_type: Optional filter for asset type.

        Returns:
            The number of assets.
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def list_versions(self, uri: AssetUri) -> List[AssetUri]:
        """
        Lists available version identifiers for a specific asset URI.

        Args:
            uri: The URI of the asset (version component is ignored).

        Returns:
            A list of URIs pointing to the specific versions of the asset.
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def is_empty(self, asset_type: str | None = None) -> bool:
        """
        Checks if the store contains any assets, optionally filtered by asset
        type.

        Args:
            asset_type: Optional filter for asset type.

        Returns:
            True if the store is empty (or empty for the given asset type),
            False otherwise.
        """
        raise NotImplementedError



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 FreeCAD Developers                                 *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the object code for the Rectangle object."""
## @package rectangle
# \ingroup draftobjects
# \brief Provides the object code for the Rectangle object.

## \addtogroup draftobjects
# @{
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
import DraftGeomUtils
from draftobjects.base import DraftObject
from draftutils import gui_utils
from draftutils import params


class Rectangle(DraftObject):
    """The Rectangle object"""

    def __init__(self, obj):
        super().__init__(obj, "Rectangle")

        _tip = QT_TRANSLATE_NOOP("App::Property", "Length of the rectangle")
        obj.addProperty("App::PropertyDistance", "Length", "Draft", _tip, locked=True)

        _tip = QT_TRANSLATE_NOOP("App::Property", "Height of the rectangle")
        obj.addProperty("App::PropertyDistance", "Height", "Draft", _tip, locked=True)

        _tip = QT_TRANSLATE_NOOP("App::Property", "Radius to use to fillet the corners")
        obj.addProperty("App::PropertyLength", "FilletRadius", "Draft", _tip, locked=True)

        _tip = QT_TRANSLATE_NOOP("App::Property", "Size of the chamfer to give to the corners")
        obj.addProperty("App::PropertyLength", "ChamferSize", "Draft", _tip, locked=True)

        _tip = QT_TRANSLATE_NOOP("App::Property", "Create a face")
        obj.addProperty("App::PropertyBool", "MakeFace", "Draft", _tip, locked=True)

        _tip = QT_TRANSLATE_NOOP("App::Property", "Horizontal subdivisions of this rectangle")
        obj.addProperty("App::PropertyInteger", "Rows", "Draft", _tip, locked=True)

        _tip = QT_TRANSLATE_NOOP("App::Property", "Vertical subdivisions of this rectangle")
        obj.addProperty("App::PropertyInteger", "Columns", "Draft", _tip, locked=True)

        _tip = QT_TRANSLATE_NOOP("App::Property", "The area of this object")
        obj.addProperty("App::PropertyArea", "Area", "Draft", _tip, locked=True)

        obj.MakeFace = params.get_param("MakeFaceMode")
        obj.Length = 1
        obj.Height = 1
        obj.Rows = 1
        obj.Columns = 1

    def onDocumentRestored(self, obj):
        super().onDocumentRestored(obj)
        gui_utils.restore_view_object(
            obj, vp_module="view_rectangle", vp_class="ViewProviderRectangle"
        )

    def execute(self, obj):
        """This method is run when the object is created or recomputed."""
        if self.props_changed_placement_only():
            obj.positionBySupport()
            self.props_changed_clear()
            return

        import Part

        if (obj.Length.Value == 0) or (obj.Height.Value == 0):
            obj.positionBySupport()
            return

        plm = obj.Placement

        shape = None

        if hasattr(obj, "Rows") and hasattr(obj, "Columns"):
            # TODO: verify if this is needed:
            if obj.Rows > 1:
                rows = obj.Rows
            else:
                rows = 1
            if obj.Columns > 1:
                columns = obj.Columns
            else:
                columns = 1
            # TODO: till here

            if (rows > 1) or (columns > 1):
                shapes = []
                l = obj.Length.Value / columns
                h = obj.Height.Value / rows
                for i in range(columns):
                    for j in range(rows):
                        p1 = App.Vector(i * l, j * h, 0)
                        p2 = App.Vector(p1.x + l, p1.y, p1.z)
                        p3 = App.Vector(p1.x + l, p1.y + h, p1.z)
                        p4 = App.Vector(p1.x, p1.y + h, p1.z)
                        p = Part.makePolygon([p1, p2, p3, p4, p1])
                        if "ChamferSize" in obj.PropertiesList:
                            if obj.ChamferSize.Value != 0:
                                w = DraftGeomUtils.filletWire(
                                    p, obj.ChamferSize.Value, chamfer=True
                                )
                                if w:
                                    p = w
                        if "FilletRadius" in obj.PropertiesList:
                            if obj.FilletRadius.Value != 0:
                                w = DraftGeomUtils.filletWire(p, obj.FilletRadius.Value)
                                if w:
                                    p = w
                        if hasattr(obj, "MakeFace"):
                            if obj.MakeFace:
                                p = Part.Face(p)
                        shapes.append(p)
                if shapes:
                    shape = Part.makeCompound(shapes)

        if not shape:
            p1 = App.Vector(0, 0, 0)
            p2 = App.Vector(p1.x + obj.Length.Value, p1.y, p1.z)
            p3 = App.Vector(p1.x + obj.Length.Value, p1.y + obj.Height.Value, p1.z)
            p4 = App.Vector(p1.x, p1.y + obj.Height.Value, p1.z)
            shape = Part.makePolygon([p1, p2, p3, p4, p1])
            if "ChamferSize" in obj.PropertiesList:
                if obj.ChamferSize.Value != 0:
                    w = DraftGeomUtils.filletWire(shape, obj.ChamferSize.Value, chamfer=True)
                    if w:
                        shape = w
            if "FilletRadius" in obj.PropertiesList:
                if obj.FilletRadius.Value != 0:
                    w = DraftGeomUtils.filletWire(shape, obj.FilletRadius.Value)
                    if w:
                        shape = w
            if hasattr(obj, "MakeFace"):
                if obj.MakeFace:
                    shape = Part.Face(shape)
            else:
                shape = Part.Face(shape)

        obj.Shape = shape

        if hasattr(obj, "Area") and hasattr(shape, "Area"):
            obj.Area = shape.Area

        obj.Placement = plm
        obj.positionBySupport()
        self.props_changed_clear()

    def onChanged(self, obj, prop):
        self.props_changed_store(prop)


# Alias for compatibility with v0.18 and earlier
_Rectangle = Rectangle

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2013 Yorik van Havre <yorik@uncreated.net>              *
# *   Copyright (c) 2019 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *   Copyright (c) 2025 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Unit tests for the Draft Workbench, SVG import and export tests."""

## @package test_svg
# \ingroup drafttests
# \brief Unit tests for the Draft Workbench, SVG import and export tests.

## \addtogroup drafttests
# @{

import os
import unittest

import FreeCAD as App
import Draft
from drafttests import auxiliary as aux
from drafttests import test_base
from draftutils.messages import _msg

try:
    import Arch
except ModuleNotFoundError:
    have_arch = False
else:
    have_arch = True


class DraftSVG(test_base.DraftTestCaseDoc):
    """Test reading and writing of SVGs with Draft."""

    def test_read_svg(self):
        """Read an SVG file and import its elements as Draft objects."""
        operation = "importSVG.import"
        _msg("  Test '{}'".format(operation))
        _msg("  This test requires an SVG file to read.")

        file = "Mod/Draft/drafttest/test.svg"
        in_file = os.path.join(App.getResourceDir(), file)
        _msg("  file={}".format(in_file))
        _msg("  exists={}".format(os.path.exists(in_file)))

        obj = aux.fake_function(in_file)
        self.assertTrue(obj, "'{}' failed".format(operation))

    def test_export_svg(self):
        """Create some figures and export them to an SVG file."""
        operation = "importSVG.export"
        _msg("  Test '{}'".format(operation))

        file = "Mod/Draft/drafttest/out_test.svg"
        out_file = os.path.join(App.getResourceDir(), file)
        _msg("  file={}".format(out_file))
        _msg("  exists={}".format(os.path.exists(out_file)))

        obj = aux.fake_function(out_file)
        self.assertTrue(obj, "'{}' failed".format(operation))

    @unittest.skipIf(not have_arch, "BIM module is not installed")
    def test_get_svg_from_arch_space_with_zero_vector(self):
        """Try to get a svg string from an Arch Space with a zero-vector as direction."""
        import Part
        import Draft

        sb = Part.makeBox(1, 1, 1)
        b = self.doc.addObject("Part::Feature", "Box")
        b.Shape = sb

        s = Arch.makeSpace(b)
        self.doc.recompute()

        try:
            Draft.get_svg(s, direction=App.Vector(0, 0, 0))
        except AttributeError as err:
            self.fail("Cryptic exception thrown: {}".format(err))
        except ValueError as err:
            App.Console.PrintLog("Exception thrown, OK: {}".format(err))
        else:
            self.fail("no exception thrown")


## @}
